"""
  Harness for visualising a neural network.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

import functools
import os
try:
  import graphviz as gv
except ImportError as ie:
  err_msg = 'To use the nn_visualise module, you need to install graphviz. This can ' + \
            'be installed via `pip install graphviz`.'
  raise ImportError('%s %s'%(err_msg, ie))

# Parameters for plotting
_SAVE_FORMAT = 'eps'
# _SAVE_FORMAT = 'png'
_LAYER_SHAPE = 'rectangle'
_IPOP_SHAPE = 'circle'
_LAYER_FONT = 'DejaVuSans'
_IPOP_FONT = 'Helvetica'
_LAYER_FONTSIZE = '16'
_FILLCOLOR = 'transparent'
_IPOP_FONTSIZE = '12'
_IPOP_FILLCOLOR = '#ffc0cb'
_DECISION_FILLCOLOR = '#98fb98'
_GRAPH_STYLES = {
  'graph': {
    'fontsize': _LAYER_FONTSIZE,
    'rankdir': 'TB',
    'label': None,
  },
  'nodes': {
  },
  'edges': {
    'arrowhead': 'open',
    'fontsize': '12',
  }
}

GV_GRAPH = functools.partial(gv.Graph, format=_SAVE_FORMAT)
GV_DIGRAPH = functools.partial(gv.Digraph, format=_SAVE_FORMAT)

# Utilities for adding nodes, edges and styles -------------------------------------------
def add_nodes(graph, nodes):
  """ Adds nodes to the graph. """
  for n in nodes:
    if isinstance(n, tuple):
      graph.node(n[0], **n[1])
    else:
      graph.node(n)
  return graph

def add_edges(graph, edges):
  """ Adds edges to the graph. """
  # pylint: disable=star-args
  for e in edges:
    if isinstance(e[0], tuple):
      graph.edge(*e[0], **e[1])
    else:
      graph.edge(*e)
  return graph

def apply_styles(graph, styles):
  """ Applies styles to the graph. """
  graph.graph_attr.update(
      ('graph' in styles and styles['graph']) or {}
  )
  graph.node_attr.update(
      ('nodes' in styles and styles['nodes']) or {}
  )
  graph.edge_attr.update(
      ('edges' in styles and styles['edges']) or {}
  )
  return graph

# Wrappers for tedious routines ----------------------------------------------------------
def _get_ip_layer(layer_idx):
  """ Returns a tuple representing the input layer. """
  return (str(layer_idx), {'label': 'i/p', 'shape': 'circle', 'style': 'filled',
                           'fillcolor': _IPOP_FILLCOLOR, 'fontsize': _IPOP_FONTSIZE,
                           'fontname': _IPOP_FONT})

def _get_op_layer(layer_idx):
  """ Returns a tuple representing the output layer. """
  return (str(layer_idx), {'label': 'o/p', 'shape': 'circle', 'style': 'filled',
                           'fillcolor':  _IPOP_FILLCOLOR, 'fontsize': _IPOP_FONTSIZE,
                           'fontname': _IPOP_FONT})

def _get_layer(layer_idx, nn, for_pres):
  """ Returns a tuple representing the layer label. """
  if nn.layer_labels[layer_idx] in ['ip', 'op']:
    fill_colour = _IPOP_FILLCOLOR
  elif nn.layer_labels[layer_idx] in ['softmax', 'linear']:
    fill_colour = _DECISION_FILLCOLOR
  else:
    fill_colour = _FILLCOLOR
  label = nn.get_layer_descr(layer_idx, for_pres)
  return (str(layer_idx), {'label': label, 'shape': 'rectangle', 'fillcolor': fill_colour,
                           'style': 'filled', 'fontname': _LAYER_FONT})

def _get_edge(layer_idx_start, layer_idx_end):
  """ Returns a tuple which is an edge. """
  return (str(layer_idx_start), str(layer_idx_end))

def _get_edges(conn_mat):
  """ Returns all edges. """
  starts, ends = conn_mat.nonzero()
  return [_get_edge(starts[i], ends[i]) for i in range(len(starts))]

# Main API ------------------------------------------------------------------------------
def visualise_nn(nn, save_file_prefix, fig_label=None, for_pres=True):
  """ The main API which will be used to visualise the network. """
  # First create nodes in the order
  nodes = [_get_layer(i, nn, for_pres) for i in range(nn.num_layers)]
  edges = _get_edges(nn.conn_mat)
  nn_graph = GV_DIGRAPH()
  add_nodes(nn_graph, nodes)
  add_edges(nn_graph, edges)
  graph_styles = _GRAPH_STYLES
  graph_styles['graph']['label'] = fig_label
  apply_styles(nn_graph, graph_styles)
  nn_graph.render(save_file_prefix)
  if os.path.exists(save_file_prefix):
    # graphviz also creates another file in the name of the prefix. delete it.
    os.remove(save_file_prefix)

def visualise_list_of_nns(list_of_nns, save_dir, fig_labels=None, fig_file_names=None,
                          for_pres=False):
  """ Visualises a list of neural networks. """
  if fig_labels is None:
    fig_labels = [None] * len(list_of_nns)
  if fig_file_names is None:
    fig_file_names = [str(idx) for idx in range(len(list_of_nns))]
  for idx, nn in enumerate(list_of_nns):
    save_file_prefix = os.path.join(save_dir, fig_file_names[idx])
    visualise_nn(nn, save_file_prefix, fig_labels[idx], for_pres)




#!/usr/bin/env python

"""
  Command line tool for Dragonfly.
  -- kandasamy@cs.cmu.edu
  -- kvysyara@andrew.cmu.edu

  Usage:
  python dragonfly-script.py --config <config file in .json or .pb format> --options <options file>
  See README in main repository for examples.
"""

# pylint: disable=relative-import
# pylint: disable=invalid-name
# pylint: disable=import-error

from __future__ import print_function
import os
# import imp
from importlib import import_module
import sys
# Local

from dragonfly import maximise_function, minimise_function, \
                      maximise_multifidelity_function, \
                      minimise_multifidelity_function, \
                      multiobjective_maximise_functions, \
                      multiobjective_minimise_functions
from dragonfly.exd.cp_domain_utils import load_config_file
from dragonfly.exd.exd_utils import get_unique_list_of_option_args
from dragonfly.utils.option_handler import get_option_specs, load_options
# Get options
from dragonfly.opt.ga_optimiser import ga_opt_args
from dragonfly.opt.gp_bandit import get_all_euc_gp_bandit_args, \
                            get_all_cp_gp_bandit_args, get_all_mf_euc_gp_bandit_args, \
                            get_all_mf_cp_gp_bandit_args
from dragonfly.opt.random_optimiser import euclidean_random_optimiser_args, \
                                   mf_euclidean_random_optimiser_args, \
                                   cp_random_optimiser_args, mf_cp_random_optimiser_args
from dragonfly.opt.multiobjective_gp_bandit import get_all_euc_moo_gp_bandit_args, \
                                                   get_all_cp_moo_gp_bandit_args
from dragonfly.opt.random_multiobjective_optimiser import \
              euclidean_random_multiobjective_optimiser_args, \
              cp_random_multiobjective_optimiser_args


dragonfly_args = [ \
  get_option_specs('config', False, None, 'Path to the json or pb config file. '),
  get_option_specs('options', False, None, 'Path to the options file. '),
  get_option_specs('max_or_min', False, 'max', 'Whether to maximise or minimise. '),
  get_option_specs('max_capital', False, -1.0,
    'Maximum capital (available budget) to be used in the experiment. '),
  get_option_specs('capital_type', False, 'return_value',
    'Maximum capital (available budget) to be used in the experiment. '),
  get_option_specs('is_multi_objective', False, 0,
    'If True, will treat it as a multiobjective optimisation problem. '),
  get_option_specs('opt_method', False, 'bo',
    ('Optimisation method. Default is bo. This should be one of bo, ga, ea, direct, ' +
     ' pdoo, or rand, but not all methods apply to all problems.')),
  get_option_specs('report_progress', False, 'default',
    ('How to report progress. Should be one of "default" (prints to stdout), ' +
     '"silent" (no reporting), or a filename (writes to file).')),
                 ]


def get_command_line_args():
  """ Returns all arguments for the command line. """
  ret = dragonfly_args + \
        ga_opt_args + \
        euclidean_random_optimiser_args + cp_random_optimiser_args + \
        mf_euclidean_random_optimiser_args + mf_cp_random_optimiser_args + \
        get_all_euc_gp_bandit_args() + get_all_cp_gp_bandit_args() + \
        get_all_mf_euc_gp_bandit_args() + get_all_mf_cp_gp_bandit_args() + \
        euclidean_random_multiobjective_optimiser_args + \
        cp_random_multiobjective_optimiser_args + \
        get_all_euc_moo_gp_bandit_args() + get_all_cp_moo_gp_bandit_args()
  return get_unique_list_of_option_args(ret)


def main():
  """ Main function. """
  options = load_options(get_command_line_args(), cmd_line=True)
  # Load domain and objective
  config = load_config_file(options.config)
  if hasattr(config, 'fidel_space'):
    is_mf = True
  else:
    is_mf = False

  # Load module
  expt_dir = os.path.dirname(os.path.abspath(os.path.realpath(options.config)))
  if not os.path.exists(expt_dir):
    raise ValueError("Experiment directory does not exist.")
  sys.path.append(expt_dir)
  obj_module = import_module(config.name, expt_dir)
  sys.path.remove(expt_dir)

  # Set capital
  if options.max_capital < 0:
    raise ValueError('max_capital (time or number of evaluations) must be positive.')

  # Call optimiser
  _print_prefix = 'Maximising' if options.max_or_min == 'max' else 'Minimising'
  call_to_optimise = {
    'single': {'max': maximise_function, 'min': minimise_function},
    'single_mf': {'max': maximise_multifidelity_function,
                  'min': minimise_multifidelity_function},
    'multi': {'max': multiobjective_maximise_functions,
              'min': multiobjective_minimise_functions},
  }
  if not options.is_multi_objective:
    if is_mf:
      print('%s multi-fidelity function on\n Fidelity-Space: %s.\n Domain: %s.\n'%(
            _print_prefix, config.fidel_space, config.domain))
      opt_val, opt_pt, history = call_to_optimise['single_mf'][options.max_or_min](
        obj_module.objective, fidel_space=None, domain=None,
        fidel_to_opt=config.fidel_to_opt, fidel_cost_func=obj_module.cost,
        max_capital=options.max_capital, capital_type=options.capital_type,
        opt_method=options.opt_method, config=config, options=options,
        reporter=options.report_progress)
    else:
      print('%s function on Domain: %s.\n'%(_print_prefix, config.domain))
      opt_val, opt_pt, history = call_to_optimise['single'][options.max_or_min](
        obj_module.objective, domain=None, max_capital=options.max_capital,
        capital_type=options.capital_type, opt_method=options.opt_method,
        config=config, options=options, reporter=options.report_progress)
    print('Optimum Value in %d evals: %0.4f'%(len(history.curr_opt_points), opt_val))
    print('Optimum Point: %s.'%(opt_pt))
  else:
    if is_mf:
      raise ValueError('Multi-objective multi-fidelity optimisation has not been ' +
                       'implemented yet.')
    else:
      print('%s %d multiobjective functions on Domain: %s.\n'%(_print_prefix,
            len(obj_module.objectives), config.domain))
      pareto_values, pareto_points, history = \
        call_to_optimise['multi'][options.max_or_min](obj_module.objectives,
        domain=None, max_capital=options.max_capital, capital_type=options.capital_type,
        opt_method=options.opt_method, config=config, options=options,
        reporter=options.report_progress)
    num_pareto_points = len(pareto_points)
    print('Found %d Pareto Points: %s.'%(num_pareto_points, pareto_points))
    print('Corresponding Pareto Values: %s.'%(pareto_values))


if __name__ == '__main__':
  sys.path.insert(0, os.getcwd())
  main()




"""
  Some generic utilities for Experiment Design.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=unbalanced-tuple-unpacking

from argparse import Namespace
import numpy as np
# Local imports
from .cp_domain_utils import sample_from_cp_domain
from ..utils.general_utils import map_to_bounds, flatten_list_of_lists
from ..utils.oper_utils import direct_ft_maximise, latin_hc_sampling, pdoo_maximise, \
                               random_maximise

# Define constants
EVAL_ERROR_CODE = 'eval_error_250320181729'

# For initialisation
# ========================================================================================
def random_sampling_cts(dim, num_samples):
  """ Just picks uniformly random samples from a  dim-dimensional space. """
  return np.random.random((num_samples, dim))

def random_sampling_kmeans_cts(dim, num_samples):
  """ Picks a large number of points uniformly at random and then runs k-means to
      select num_samples points. """
  try:
    from sklearn.cluster import KMeans
    num_candidates = np.clip(100*(dim**2), 4*num_samples, 20*num_samples)
    candidates = random_sampling_cts(dim, num_candidates)
    centres = KMeans(n_clusters=num_samples).fit(candidates)
    return centres.cluster_centers_
  except ImportError:
    return random_sampling_cts(dim, num_samples)

def random_sampling_discrete(param_values, num_samples):
  """ Picks uniformly random samples from a discrete set of values.
      param_values is a list of lists where each internal list is the set of values
      for each parameter. """
  num_params = len(param_values)
  ret_vals_for_each_param = []
  for i in range(num_params):
    rand_idxs = np.random.randint(0, len(param_values[i]), size=(num_samples,))
    ret_vals_for_each_param.append([param_values[i][idx] for idx in rand_idxs])
  # Now put them into an num_samples x num_params array
  ret = []
  for j in range(num_samples):
    curr_sample = [ret_vals_for_each_param[param_idx][j] for param_idx in
                                                             range(num_params)]
    ret.append(curr_sample)
  return ret


# A wrapper to get initial points ------------------------------------------
def get_euclidean_initial_points(init_method, num_samples, domain_bounds):
  """ Gets the initial set of points for a Euclidean space depending on the init_method.
  """
  dim = len(domain_bounds)
  if init_method == 'rand':
    ret = random_sampling_cts(dim, num_samples)
  elif init_method == 'rand_kmeans':
    ret = random_sampling_kmeans_cts(dim, num_samples)
  elif init_method == 'latin_hc':
    ret = latin_hc_sampling(dim, num_samples)
  else:
    raise ValueError('Unknown init method %s.'%(init_method))
  return map_to_bounds(ret, domain_bounds)


def _process_fidel_for_initialisation(fidel, fidel_to_opt, set_to_fidel_to_opt_with_prob):
  """ Returns fidel to fidel_to_opt with probability set_to_fidel_to_opt_with_prob. """
  if np.random.random() < set_to_fidel_to_opt_with_prob:
    return fidel_to_opt
  else:
    return fidel

def get_euclidean_initial_fidels(init_method, num_samples, fidel_space_bounds,
                                 fidel_to_opt, set_to_fidel_to_opt_with_prob=None):
  """ Returns an initial set of fidels for Euclidean spaces depending on the init_method.
      sets the fidel to fidel_to_opt with probability set_to_fidel_to_opt_with_prob.
  """
  # An internal function to process fidelity
  set_to_fidel_to_opt_with_prob = 0.0 if set_to_fidel_to_opt_with_prob is None else \
                                  set_to_fidel_to_opt_with_prob
  init_fidels = get_euclidean_initial_points(init_method, num_samples, fidel_space_bounds)
  return [_process_fidel_for_initialisation(fidel, fidel_to_opt,
                                            set_to_fidel_to_opt_with_prob)
          for fidel in init_fidels]


def get_cp_domain_initial_fidels(fidel_space, num_samples, fidel_to_opt,
                                 set_to_fidel_to_opt_with_prob,
                                 euclidean_sample_type='latin_hc',
                                 integral_sample_type='latin_hc',
                                 nn_sample_type='rand'):
  """ Function to return the initial fidelities for CP Domain. """
  set_to_fidel_to_opt_with_prob = 0.0 if set_to_fidel_to_opt_with_prob is None else \
                                  set_to_fidel_to_opt_with_prob
  init_fidels = sample_from_cp_domain(fidel_space, num_samples,
                              euclidean_sample_type=euclidean_sample_type,
                              integral_sample_type=integral_sample_type,
                              nn_sample_type=nn_sample_type)
  return [_process_fidel_for_initialisation(fidel, fidel_to_opt,
                                            set_to_fidel_to_opt_with_prob)
          for fidel in init_fidels]


def get_euclidean_initial_qinfos(domain_init_method, num_samples, domain_bounds,
                                 fidel_init_method=None, fidel_space_bounds=None,
                                 fidel_to_opt=None, set_to_fidel_to_opt_with_prob=None):
  """ Returns the initial points in qinfo Namespaces. """
  init_points = get_euclidean_initial_points(domain_init_method, num_samples,
                                             domain_bounds)
  if fidel_space_bounds is None:
    return [Namespace(point=init_point) for init_point in init_points]
  else:
    init_fidels = get_euclidean_initial_fidels(fidel_init_method, num_samples,
                   fidel_space_bounds, fidel_to_opt, set_to_fidel_to_opt_with_prob)
    return [Namespace(point=ipt, fidel=ifl) for (ipt, ifl)
            in zip(init_points, init_fidels)]


def get_cp_domain_initial_qinfos(domain, num_samples, fidel_space=None, fidel_to_opt=None,
                                 set_to_fidel_to_opt_with_prob=None,
                                 dom_euclidean_sample_type='latin_hc',
                                 dom_integral_sample_type='latin_hc',
                                 dom_nn_sample_type='rand',
                                 fidel_space_euclidean_sample_type='latin_hc',
                                 fidel_space_integral_sample_type='latin_hc',
                                 fidel_space_nn_sample_type='rand'):
  """ Get initial qinfos in Cartesian product domain. """
  # pylint: disable=too-many-arguments
  ret_dom_pts = sample_from_cp_domain(domain, num_samples,
                                      euclidean_sample_type=dom_euclidean_sample_type,
                                      integral_sample_type=dom_integral_sample_type,
                                      nn_sample_type=dom_nn_sample_type)
  if fidel_space is None:
    ret_dom_pts = ret_dom_pts[:num_samples]
    return [Namespace(point=x) for x in ret_dom_pts]
  else:
    ret_fidels = get_cp_domain_initial_fidels(fidel_space, num_samples, fidel_to_opt,
                   set_to_fidel_to_opt_with_prob,
                   euclidean_sample_type=fidel_space_euclidean_sample_type,
                   integral_sample_type=fidel_space_integral_sample_type,
                   nn_sample_type=fidel_space_nn_sample_type)
    return [Namespace(point=ipt, fidel=ifl) for (ipt, ifl) in
            zip(ret_dom_pts, ret_fidels)]

# Maximise with method here.
# Some wrappers for all methods
def maximise_with_method(method, obj, domain, max_evals, return_history=False,
                         *args, **kwargs):
  """ A wrapper which optimises obj over the domain domain. """
  # If the method itself is a function, just use it.
  if hasattr(method, '__call__'):
    return method(obj, domain, max_evals, return_history, *args, **kwargs)
  # The common use case is that the method is a string.
  if domain.get_type() == 'euclidean':
    return maximise_with_method_on_euclidean_domain(method, obj, domain.bounds, max_evals,
                                                    return_history, *args, **kwargs)
  elif domain.get_type() == 'integral':
    return maximise_with_method_on_integral_domain(method, obj, domain.bounds, max_evals,
                                                   domain.get_dim(), return_history,
                                                   *args, **kwargs)
  elif domain.get_type() == 'prod_discrete':
    return maximise_with_method_on_prod_discrete_domain(method, obj, domain, max_evals,
                                                        return_history, *args, **kwargs)
  elif domain.get_type() == 'cartesian_product':
    return maximise_with_method_on_cp_domain(method, obj, domain, max_evals,
                                             return_history, *args, **kwargs)
  else:
    raise ValueError('Unknown domain type %s.'%(domain.get_type()))


def maximise_with_method_on_euclidean_domain(method, obj, bounds, max_evals, dim,
                                             return_history=False, *args, **kwargs):
  """ A wrapper for euclidean spaces which calls one of the functions below based on the
      method. """
  if method.lower().startswith('rand'):
    max_val, max_pt, history = \
      random_maximise(obj, bounds, max_evals, return_history, *args, **kwargs)
  elif method.lower().startswith('direct') and dim <= 60:
    max_val, max_pt, history = \
      direct_ft_maximise(obj, bounds, max_evals, return_history, *args, **kwargs)
  elif method.lower().startswith('pdoo') or (method.lower().startswith('direct')
                                             and dim > 60):
    max_val, max_pt, history = \
      pdoo_maximise(obj, bounds, max_evals, *args, **kwargs)
  else:
    raise ValueError('Unknown maximisation method: %s.'%(method))
  if return_history:
    return max_val, max_pt, history
  else:
    return max_val, max_pt

def maximise_with_method_on_integral_domain(method, obj, bounds, max_evals,
                                            return_history=False, *args, **kwargs):
  """ A wrapper for integral spaces which calls one of the functions below based on the
      method. """
  # pylint: disable=unused-argument
  raise NotImplementedError('Not implemented integral domain optimisers yet.')


def maximise_with_method_on_prod_discrete_domain(method, obj, domain, max_evals,
                                                 return_history=False, *args, **kwargs):
  """ A wrapper for discrete spaces which calls one of the functions below based on the
      method. """
  # pylint: disable=unused-argument
  raise NotImplementedError('Not implemented discrete domain optimisers yet.')


def maximise_with_method_on_product_euclidean_spaces(method, obj, list_of_euc_domains,
                                                     max_evals, return_history=False,
                                                     *args, **kwargs):
  """ Maximises an objective with method on a product Euclidean space. """
  def _regroup_flattended_point(pt, _dom_dims, _cum_dims):
    """ Regroups a flattened point into a list of lists. """
    ret = [pt[cd:cd+d] for (cd, d) in zip(_cum_dims, _dom_dims)]
    assert len(ret) == len(_dom_dims) # Check if dimensions match.
    return ret
  dom_dims = [dom.dim for dom in list_of_euc_domains]
  cum_dims = list(np.cumsum(dom_dims))
  cum_dims = [0] + cum_dims[:-1]
  euc_dom_bounds = flatten_list_of_lists([dom.bounds for dom in list_of_euc_domains])
  modified_obj = lambda x: obj(_regroup_flattended_point(x, dom_dims, cum_dims))
  opt_result = maximise_with_method_on_euclidean_domain(method, modified_obj,
    euc_dom_bounds, max_evals, return_history, *args, **kwargs)
  if return_history:
    max_val, max_pt, history = opt_result
  else:
    max_val, max_pt = opt_result
  # regroup the maximum point
  regrouped_max_pt = _regroup_flattended_point(max_pt, dom_dims, cum_dims)
  if return_history:
    return max_val, regrouped_max_pt, history
  else:
    return max_val, regrouped_max_pt


def _rand_maximise_vectorised_objective_in_cp_domain(obj, domain, max_evals,
  return_history=False):
  """ Maximises a vectorised function in Cartesian product spaces.
      Mostly used for TS style acquisitions in BO.
  """
  rand_samples = sample_from_cp_domain(domain, int(max_evals))
  rand_values = [obj(x) for x in rand_samples]
  max_idx = np.argmax(rand_values)
  max_pt = rand_samples[max_idx]
  max_val = rand_values[max_idx]
  if return_history:
    history = Namespace(query_points=rand_samples,
                        query_vals=rand_values)
    return max_val, max_pt, history
  else:
    return max_val, max_pt


def maximise_with_method_on_cp_domain(method, obj, domain, max_evals,
                                      return_history=False, *args, **kwargs):
  """ A wrapper for maximising an objective on a CartesianProductDomain. """
  # pylint: disable=too-many-locals
  # pylint: disable=too-many-branches
  if method.lower().startswith(('direct', 'pdoo')): # ------------------------------------
    # If the domain consists entirely of Euclidean parts, use PDOO or DiRect
    return maximise_with_method_on_product_euclidean_spaces(method, obj,
      domain.list_of_domains, max_evals, return_history, *args, **kwargs)
  elif method.lower() == 'rand': # -------------------------------------------------------
    # Use a random optimiser
    return _rand_maximise_vectorised_objective_in_cp_domain(obj, domain, max_evals,
                                                            return_history)
  elif method.lower().startswith('ga'): # ------------------------------------------------
    # First decide if there is a follow up Euclidean opt method
    ga_opt_methods = method.lower().split('-')
    euc_domains = [dom for dom in domain.list_of_domains if dom.get_type() == 'euclidean']
    if len(ga_opt_methods) == 2 and len(euc_domains) > 0:
      to_follow_up_with_euc_opt = True
      euc_opt_method = ga_opt_methods[1]
      euc_domain_idxs = [idx for (idx, dom) in enumerate(domain.list_of_domains) if
                         dom.get_type() == 'euclidean']
    else:
      to_follow_up_with_euc_opt = False
    # Run a GA optimiser ---------------------------------------
    from .worker_manager import SyntheticWorkerManager
    from .experiment_caller import CPFunctionCaller
    from ..opt.cp_ga_optimiser import cp_ga_optimiser_from_proc_args
    obj_in_func_caller = CPFunctionCaller(obj, domain, domain_orderings=None)
    worker_manager = SyntheticWorkerManager(1, time_distro='const')
    ga_max_val, ga_max_pt, ga_history = cp_ga_optimiser_from_proc_args(obj_in_func_caller,
      domain, worker_manager, max_evals, mode='asy', options=None, reporter='silent')
    # Finished running a GA optimiser --------------------------
    if to_follow_up_with_euc_opt:
      # If you need to do an additional optimisation on the Euclidean part, do so.
      def _swap_indices_at_point(swap_pt, orig_pt, swap_idxs):
        """ Swap indices at point swap_pt. """
        ret = orig_pt[:]
        for sidx, spt in zip(swap_idxs, swap_pt):
          ret[sidx] = spt
        return ret
      def _get_swapped_obj(_obj, _orig_pt, _swap_idxs):
        """ Returns an objective which swaps in new points at _swap_idxs. """
        return lambda x: _obj(_swap_indices_at_point(x, _orig_pt, _swap_idxs))
      swapped_obj = _get_swapped_obj(obj, ga_max_pt, euc_domain_idxs)
      euc_max_val, euc_max_pt = \
        maximise_with_method_on_product_euclidean_spaces(euc_opt_method, swapped_obj,
        euc_domains, max_evals, return_history=False)
      if euc_max_val > ga_max_val:
        max_val = euc_max_val
        history = ga_history
        max_pt = _swap_indices_at_point(euc_max_pt, ga_max_pt, euc_domain_idxs)
      else:
        max_val, max_pt, history = (ga_max_val, ga_max_pt, ga_history)
    else:
      max_val, max_pt, history = (ga_max_val, ga_max_pt, ga_history)
    # return
    if return_history:
      return max_val, max_pt, history
    else:
      return max_val, max_pt
  else:
    raise NotImplementedError('Not implemented yet!')


# Miscellaneous
# ======================================================================
def get_unique_list_of_option_args(all_args):
  """ Returns a unique list of option args. """
  ret = []
  ret_names = []
  for arg in all_args:
    if arg['name'] not in ret_names:
      ret.append(arg)
      ret_names.append(arg['name'])
  return ret




"""
  Some utilities for graph operations.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import


from copy import deepcopy
import numpy as np
from scipy.linalg import eigh as sp_eigh
# Local imports
from .general_utils import get_nonzero_indices_in_vector


def construct_random_adjacency_matrix(num_vertices, avg_edge_frac=0.4, directed=False):
  """ Constructs a random adjacency matrix where avg_edge_frac of the edges are connected
      on average. """
  M = np.random.random((num_vertices, num_vertices))
  threshold = avg_edge_frac if directed else 2 * avg_edge_frac
  M = M if directed else (M + M.T)
  A = (M < threshold).astype(np.float)
  np.fill_diagonal(A, 0)
  return A


def get_undirected_adj_mat_from_directed(directed_adj_mat):
  """ Given a directed adjacency matrix, returns the undirected matrix. """
  num_nodes = directed_adj_mat.shape[0]
  ret = directed_adj_mat + directed_adj_mat.T
  ret[np.diag_indices(num_nodes)] = directed_adj_mat[np.diag_indices(num_nodes)]
  return ret


def is_valid_adj_mat(A):
  """ Returns true if A is a valid adjacency matrix. i.e. it is symmetric and all
      diagonals are 0. """
  return (A == A.T).all() and (A.diagonal() == 0).all()


def get_normalised_laplacian(A):
  """ Returns the (Symmetric) Normalised Graph Laplacian from an adjacency matrix A. """
  assert is_valid_adj_mat(A)
  with np.errstate(divide='ignore'):
    D_sqrt_inv = 1./ np.sqrt(A.sum(axis=0))
    D_sqrt_inv[D_sqrt_inv == np.inf] = 0
  L = np.eye(A.shape[0]) - (A * D_sqrt_inv) * D_sqrt_inv[:, np.newaxis]
  return L


def get_smallest_symm_eig(M, smallest_eigval_idx=None):
  """ Returns the smallest_eigval_idx eigenvalues and eigenvectors of the symmetric
      matrix M. """
  smallest_eigval_idx = (M.shape[0] - 1) if smallest_eigval_idx is None \
                        else smallest_eigval_idx
  eig_val_indices = [0, min(smallest_eigval_idx, M.shape[0]-1)]
  eigvals, eigvecs = sp_eigh(M, eigvals=eig_val_indices)
  return eigvals, eigvecs


def compute_lap_embedding_from_laplacian(L, num_proj_dimensions=None,
                                         proj_to_sphere=True):
  """ Returns the representation using Laplacian eigenvectors and in the unit cube. """
  num_proj_dimensions = num_proj_dimensions if num_proj_dimensions is not None \
                        else L.shape[0] # return entire matrix.
  _, eigvecs = get_smallest_symm_eig(L, num_proj_dimensions-1)
  if proj_to_sphere:  # each row has unit euclidean norm
    row_wise_norms = np.sqrt(np.sum(eigvecs**2, axis=1))
    sphere_repr = eigvecs / row_wise_norms[:, np.newaxis]
  # Pad with zeros
  if sphere_repr.shape[1] < num_proj_dimensions:
    zero_mat = np.zeros((L.shape[0], num_proj_dimensions - sphere_repr.shape[1]))
    sphere_repr = np.concatenate((sphere_repr, zero_mat), axis=1)
  return sphere_repr


def compute_lap_embedding_from_graph(A, num_proj_dimensions=None, proj_to_sphere=True):
  """ Returns the representation using Laplacian eigenvectors and in the unit cube.
      Computes this from the adjacency matrix by first computing the Laplacian."""
  L = get_normalised_laplacian(A)
  return compute_lap_embedding_from_laplacian(L, num_proj_dimensions, proj_to_sphere)


def apsp_floyd_warshall_costs(A):
  """ Runs the Floyd Warshall algorithm to return an nxn matrix which computes the
      all pairs shortest paths. Here A(i,j) is treated as the distance from i to j.
      So 0's will be counted as 0's. Non-edges should be specified as infinities.
      Just copying the pseudo code in Wikipedia."""
  dist = deepcopy(A)
  np.fill_diagonal(dist, 0)
  n = A.shape[0]
  for k in range(n):
    for i in range(n):
      for j in range(n):
        if dist[i, j] > dist[i, k] + dist[k, j]:
          dist[i, j] = dist[i, k] + dist[k, j]
  return dist


def apsp_floyd_warshall(A):
  """ Runs the Floyd Warshall algorithm to return an nxn matrix which computes the
      all pairs shortest paths. Here 0's denote non-edges. """
  dist = deepcopy(A)
  # Create an initial matrix with infinity for all non-edges and zero on the diagonals.
  dist[dist == 0] = np.inf
  np.fill_diagonal(dist, 0)
  return apsp_floyd_warshall_costs(dist)


def get_children(node_idx, A):
  """ Returns the children of node_idx according to the adjacency matrix A. """
  return get_nonzero_indices_in_vector(A[node_idx])

def get_parents(node_idx, A):
  """ Returns the parents of node_idx according to the adjacency matrix A. """
  return get_nonzero_indices_in_vector(A[:, node_idx])


def apsp_dijkstra(A):
  """ Runs Dijkstra's on all nodes to compute the all pairs shortest paths. """
  vertex_dists = []
  for vertex in range(A.shape[0]):
    curr_vertex_dists = dijkstra(A, vertex)
    vertex_dists.append(curr_vertex_dists)
  return np.array(vertex_dists)


def dijkstra(A, source, non_edges_are_zero_or_inf='zero'):
  """ Run's dijkstra's on the vertex to produce the shortest path to all nodes.
      Just copyng the pseudo code in Wikipedia.
      non_edges_are_zero_or_inf indicate whether a non-edge is indicated as a 0 or
      inf in A.
  """
  vertex_is_remaining = np.array([1] * A.shape[0])
  all_vertices = np.array(range(A.shape[0]))
  all_dists = np.array([np.inf] * A.shape[0])
  all_dists[source] = 0
  while sum(vertex_is_remaining) > 0:
    # Find the minimum and remove it.
    rem_dists = deepcopy(all_dists)
    rem_dists[vertex_is_remaining == 0] = np.nan
    u = np.nanargmin(rem_dists)
    vertex_is_remaining[u] = 0
    if np.all(np.logical_not(np.isfinite(rem_dists))):
      break
    # Now apply dijkstra's updates
    if non_edges_are_zero_or_inf == 'zero':
      u_nbd = all_vertices[A[u] > 0]
    elif non_edges_are_zero_or_inf == 'inf':
      u_nbd = all_vertices[A[u] < np.inf]
    else:
      raise ValueError('non_edges_are_zero_or_inf should be \'zero\' or \'inf\'.')
    for v in u_nbd:
      alt = all_dists[u] + A[u][v]
      if alt < all_dists[v]:
        all_dists[v] = alt
  return all_dists


def compute_nn_path_lengths(A, top_order, path_type):
  """ Computes the path lengths on a NN with adjacency matrix A to top_order[-1].
      top_order is a topological ordering of the nodes.
      A(i,j) is finite means there is an edge between i and j.
  """
  if path_type in 'shortest':
    get_curr_length_from_child_lengths = min
  elif path_type == 'longest':
    get_curr_length_from_child_lengths = max
  elif path_type == 'rw':
    get_curr_length_from_child_lengths = lambda x: sum(x) / float(len(x))
  else:
    raise ValueError('Unknown path_type: %s.'%(path_type))
  # Now the algorithm
  all_vertices = np.array(range(A.shape[0]))
  source = top_order[-1]
  all_dists = np.array([np.inf] * A.shape[0])
  all_dists[source] = 0
  for layer_idx in reversed(top_order[:-1]):
    curr_children = all_vertices[A[layer_idx] < np.inf]
    children_path_lengths = [all_dists[ch_idx] for ch_idx in curr_children]
    children_path_lengths = [children_path_lengths[i] + A[layer_idx, curr_children[i]] for
                             i in range(len(curr_children))]
    all_dists[layer_idx] = get_curr_length_from_child_lengths(children_path_lengths)
  return all_dists


def kahn_topological_sort(A, start_nodes):
  """ Applies Kahn's algorithm to return a topological sort of the graph. Starts with
      start_nodes. A is adjacency graph. Following pseudo code in Wikipedia. """
  ret = []
  edges = deepcopy(A)
  S = deepcopy(start_nodes) if hasattr(start_nodes, '__iter__') else [start_nodes]
  while len(S) > 0:
    curr_node = S.pop(0)
    ret.append(curr_node)
    curr_nbd = get_nonzero_indices_in_vector(edges[curr_node])
    edges[curr_node] = 0
    for neighbor in curr_nbd:
      if not edges[:, neighbor].sum() > 0:
        S.append(neighbor)
  has_cycles = False if edges.sum() == 0 else True
  return ret, has_cycles




"""
  Unit tests for continuous distributions.
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
import unittest
# pylint: disable=no-self-use

import numpy as np
import warnings

# Local imports
from .continuous import Beta, ContinuousUniform, Exponential, \
                                     Normal, MultivariateGaussian
from ..utils.base_test_class import BaseTestClass, execute_tests

class ContinuousDistributionsTestCase(BaseTestClass):
  """ Unit tests for distributions in continuous.py """

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 1000000
    self.threshold = 0.01
    warnings.filterwarnings("ignore", message="numpy.dtype size changed")

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples, axis=None):
    """ Computes Mean """
    return np.mean(samples, axis)

  def _compute_variance(self, samples, axis=None):
    """ Computes Variance """
    return np.var(samples, axis)

  def _compute_covariance(self, samples):
    """ Computes Covariance """
    return np.cov(samples.T)

  def test_rand_sampling_normal(self):
    """ Tests random sampling from Normal distribution """
    self.report('Test random sampling of Normal Distribution.')
    mean = 0
    variance = 1
    dist = Normal(mean, variance)
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(variance - var_r) <= self.threshold
    self.report('%s :: test result: mean=%0.3f, variance=%0.3f'%\
                (str(dist), mean_r, var_r), 'test_result')

  def test_rand_sampling_multi_normal(self):
    """ Tests random sampling from Multivariate Normal distribution """
    self.report('Test random sampling of Multivariate Normal Distribution.')
    cov_thresh = 0.1
    mean_thresh = 0.01
    mean = np.arange(3)
    covariance = 3*np.identity(3)
    dist = MultivariateGaussian(mean, covariance)
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples, 0)
    self._check_sample_sizes(samples)
    assert (abs(mean - self._compute_mean(samples, 0)) <= mean_thresh).all()
    assert (abs(covariance - self._compute_covariance(samples)) <= cov_thresh).all()
    self.report('%s :: test result: mean=%s'%(str(dist), str(mean_r)), 'test_result')

  def test_rand_sampling_expo(self):
    """ Tests random sampling from Exponential distribution """
    self.report('Test random sampling of Exponential Distribution.')
    lam = 2
    dist = Exponential(lam)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  @unittest.skip("Stochastic")
  def test_rand_sampling_uniform(self):
    """ Tests random sampling from Continuous Uniform """
    self.report('Test random sampling of Continuous Uniform Distribution.')
    lower = -5
    upper = 5
    dist = ContinuousUniform(lower, upper)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  def test_rand_sampling_beta(self):
    """ Tests random sampling from Beta Distribution """
    self.report('Test random sampling of Beta Distribution.')
    alpha = 1
    beta = 2
    dist = Beta(alpha, beta)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('random', self.size)
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()



"""
  A module for fitting a GP and tuning its kernel.
  -- kandasamy@cs.cmu.edu
  -- kvysyara@andrew.cmu.edu
"""

from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=no-member
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-statements

from argparse import Namespace
from itertools import product as itertools_product
import sys
import numpy as np
# Local imports
from ..distributions.model import Model
from ..distributions import continuous
from ..distributions import discrete
from ..utils.general_utils import stable_cholesky, draw_gaussian_samples, \
     project_symmetric_to_psd_cone, solve_lower_triangular, solve_upper_triangular
from ..utils.oper_utils import direct_ft_maximise, pdoo_maximise, \
                             random_maximise, random_sample_cts_dscr
from ..utils.option_handler import get_option_specs, load_options
from ..utils.reporters import get_reporter


# These are mandatory requirements. Every GP implementation should probably use them.
mandatory_gp_args = [ \
  get_option_specs('hp_tune_criterion', False, 'ml',
                   'Which criterion to use when tuning hyper-parameters. Other ' +
                   'options are post_sampling and post_mean.'),
  get_option_specs('hp_tune_probs', False, 'uniform', \
    'With what probability should we choose each strategy given in hp_tune_criterion.' + \
    'If "uniform" we we will use uniform probabilities and if "adaptive" we will use ' + \
    'adaptive probabilities which weight acquisitions according to how well they do.'),
  get_option_specs('ml_hp_tune_opt', False, 'default',
                   'Which optimiser to use when maximising the tuning criterion.'),
  get_option_specs('hp_tune_max_evals', False, -1,
                   'How many evaluations to use when maximising the tuning criterion.'),
  get_option_specs('handle_non_psd_kernels', False, 'guaranteed_psd',
                   'How to handle kernels that are non-psd.'),
  # The mean and noise variance of the GP
  get_option_specs('mean_func_type', False, 'tune',
                   ('Specify the type of mean function. Should be mean, median, const ',
                    'zero, or tune. If const, specifcy value in mean-func-const.')),
  get_option_specs('mean_func_const', False, 0.0,
                   'The constant value to use if mean_func_type is const.'),
  get_option_specs('noise_var_type', False, 'tune', \
    ('Specify how to obtain the noise variance. Should be tune, label or value. ' \
     'Specify appropriate value in noise_var_label or noise_var_value')),
  get_option_specs('noise_var_label', False, 0.05,
                   'The fraction of label variance to use as noise variance.'),
  get_option_specs('noise_var_value', False, 0.1,
                   'The (absolute) value to use as noise variance.'),
  get_option_specs('post_hp_tune_method', False, 'slice',
                   'Which sampling to use when maximising the tuning criterion. Other ' +
                   'option is nuts.'),
  get_option_specs('post_hp_tune_burn', False, -1,
                   'How many initial samples to ignore during sampling.'),
  get_option_specs('post_hp_tune_offset', False, 25,
                   'How many samples to ignore between samples.'), \
  ]


def _check_feature_label_lengths_and_format(X, Y):
  """ Checks if the length of X and Y are the same. """
  if len(X) != len(Y):
    raise ValueError('Length of X (' + len(X) + ') and Y (' + \
      len(Y) + ') do not match.')

def _get_default_ml_hp_tune_opt(num_hps):
  """ Return default optimisation method for maximum likelihood. """
  if num_hps > 60:
    return 'pdoo'
  else:
    return 'direct'


# Main GP class -------------------------------------------------------
class GP(object):
  """ Base class for Gaussian processes. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, X, Y, kernel, mean_func, noise_var, build_posterior=True,
               reporter=None, handle_non_psd_kernels='guaranteed_psd'):
    """ Constructor. """
    super(GP, self).__init__()
    _check_feature_label_lengths_and_format(X, Y)
    self.set_data(X, Y, build_posterior=False)
    self.kernel = kernel
    self.mean_func = mean_func
    self.noise_var = noise_var
    self.reporter = get_reporter(reporter)
    self.handle_non_psd_kernels = handle_non_psd_kernels
    # Some derived attribues.
    self.num_tr_data = len(self.Y)
    # Initialise other attributes we will need.
    self.L = None
    self.alpha = None
    self.K_trtr_wo_noise = None
    self._set_up()
    # Build posterior if necessary
    if build_posterior:
      self.build_posterior()

  def _set_up(self):
    """ Additional set up. """
    # Test that if the kernel is not guaranteed to be PSD, then have have a way to
    # handle it.
    if not self.kernel.is_guaranteed_psd():
      assert self.handle_non_psd_kernels in ['project_first',
                                             'try_before_project']

  def _write_message(self, msg):
    """ Writes a message via the reporter or the std out. """
    if self.reporter:
      self.reporter.write(msg)
    else:
      sys.stdout.write(msg)

  def set_data(self, X, Y, build_posterior=True):
    """ Sets the data to X and Y. """
    self.X = list(X)
    self.Y = list(Y)
    self.num_tr_data = len(self.Y)
    if build_posterior:
      self.build_posterior()

  def add_data_single(self, x_new, y_new, *args, **kwargs):
    """ Adds a single data point. """
    self.add_data_multiple([x_new], [y_new], *args, **kwargs)

  def add_data_multiple(self, X_new, Y_new, build_posterior=True):
    """ Adds new data to the GP. If rebuild is true it rebuilds the posterior. """
    _check_feature_label_lengths_and_format(X_new, Y_new)
    self.X.extend(X_new)
    self.Y.extend(Y_new)
    self.num_tr_data = len(self.Y)
    if build_posterior:
      self.build_posterior()

  # Methods for computing the posterior ----------------------------------------------
  def _get_training_kernel_matrix(self):
    """ Returns the training kernel matrix. Writing this as a separate method in case,
        the kernel computation can be done efficiently for a child class.
    """
    return self.kernel(self.X, self.X)

  def build_posterior(self):
    """ Builds the posterior GP by computing the mean and covariance. """
    K_trtr_wo_noise = self._get_training_kernel_matrix()
    self.K_trtr_wo_noise = K_trtr_wo_noise
    self.L = _get_cholesky_decomp(K_trtr_wo_noise, self.noise_var,
                                  self.handle_non_psd_kernels)
    Y_centred = self.Y - self.mean_func(self.X)
    self.alpha = solve_upper_triangular(self.L.T,
                                        solve_lower_triangular(self.L, Y_centred))

  def eval(self, X_test, uncert_form='none'):
    """ Evaluates the GP on X_test. If uncert_form is
          covar: returns the entire covariance on X_test (nxn matrix)
          std: returns the standard deviations on the test set (n vector)
          none: returns nothing (default).
    """
    # Compute the posterior mean.
    test_mean = self.mean_func(X_test)
    K_tetr = self.kernel(X_test, self.X)
    pred_mean = test_mean + K_tetr.dot(self.alpha)
    # Compute the posterior variance or standard deviation as required.
    if uncert_form == 'none':
      uncert = None
    else:
      K_tete = self.kernel(X_test, X_test)
      V = solve_lower_triangular(self.L, K_tetr.T)
      post_covar = K_tete - V.T.dot(V)
      post_covar = get_post_covar_from_raw_covar(post_covar, self.noise_var,
                                                 self.kernel.is_guaranteed_psd())
      if uncert_form == 'covar':
        uncert = post_covar
      elif uncert_form == 'std':
        uncert = np.sqrt(np.diag(post_covar))
      else:
        raise ValueError('uncert_form should be none, covar or std.')
    return pred_mean, uncert

  def eval_with_hallucinated_observations(self, X_test, X_halluc, uncert_form='none'):
    """ Evaluates the GP with additional hallucinated observations in the
        kernel matrix. """
    pred_mean, _ = self.eval(X_test, uncert_form='none') # Just compute the means.
    if uncert_form == 'none':
      uncert = None
    else:
      # Computed the augmented kernel matrix and its cholesky decomposition.
      X_aug = self.X + X_halluc
      K_haha = self.kernel(X_halluc, X_halluc) # kernel for hallucinated data
      K_trha = self.kernel(self.X, X_halluc) # kernel with training and hallucinated data
      aug_K_trtr_wo_noise = np.vstack((np.hstack((self.K_trtr_wo_noise, K_trha)),
                                       np.hstack((K_trha.T, K_haha))))
      aug_L = _get_cholesky_decomp(aug_K_trtr_wo_noise, self.noise_var,
                                   self.handle_non_psd_kernels)
      # Augmented kernel matrices for the test data
      aug_K_tete = self.kernel(X_test, X_test)
      aug_K_tetr = self.kernel(X_test, X_aug)
      aug_V = solve_lower_triangular(aug_L, aug_K_tetr.T)
      aug_post_covar = aug_K_tete - aug_V.T.dot(aug_V)
      aug_post_covar = get_post_covar_from_raw_covar(aug_post_covar, self.noise_var,
                                                     self.kernel.is_guaranteed_psd())
      if uncert_form == 'covar':
        uncert = aug_post_covar
      elif uncert_form == 'std':
        uncert = np.sqrt(np.diag(aug_post_covar))
      else:
        raise ValueError('uncert_form should be none, covar or std.')
    return (pred_mean, uncert)

  def compute_log_marginal_likelihood(self):
    """ Computes the log marginal likelihood. """
    Y_centred = self.Y - self.mean_func(self.X)
    ret = -0.5 * Y_centred.T.dot(self.alpha) - (np.log(np.diag(self.L))).sum() \
          - 0.5 * self.num_tr_data * np.log(2*np.pi)
    return ret

  def compute_grad_log_marginal_likelihood(self, param, *args):
    """ Computes the gradient of log marginal likelihood. """
    alpha = np.expand_dims(self.alpha, axis=0)
    if param == 'noise_var':
      grad_m = self.noise_var*np.identity(len(self.X))
    elif param == 'noise_mean':
      return np.asscalar(np.matmul(alpha, np.ones((len(self.Y), 1))))
    else:
      grad_m = self.kernel.gradient(param, self.X, self.X, *args)
    grad_m = np.matmul(alpha.T, np.matmul(alpha, grad_m)) - \
             solve_upper_triangular(self.L.T, solve_lower_triangular(self.L, grad_m))
    return 0.5 * np.trace(grad_m)

  def __str__(self):
    """ Returns a string representation of the GP. """
    return '%s, eta2=%0.3f (n=%d)'%(self._child_str(), self.noise_var, len(self.Y))

  def _child_str(self):
    """ String representation for child GP. """
    raise NotImplementedError('Implement in child class. !')

  def draw_samples(self, num_samples, X_test=None, mean_vals=None, covar=None):
    """ Draws num_samples samples at returns their values at X_test. """
    if X_test is not None:
      mean_vals, covar = self.eval(X_test, 'covar')
    return draw_gaussian_samples(num_samples, mean_vals, covar)

  def draw_samples_with_hallucinated_observations(self, num_samples, X_test,
                                                  X_halluc):
    """ Draws samples with hallucinated observations. """
    mean_vals, aug_covar = self.eval_with_hallucinated_observations(X_test, \
                                             X_halluc, uncert_form='covar')
    return draw_gaussian_samples(num_samples, mean_vals, aug_covar)

  def visualise(self, file_name=None, boundary=None, true_func=None,
                num_samples=20, conf_width=3):
    """ Visualises the GP. """
    # pylint: disable=unused-variable
    # pylint: disable=too-many-locals
    if hasattr(self.kernel, 'dim') and self.kernel.dim != 1:
      self._write_message('Cannot visualise only in 1 dimension.\n')
    else:
      import matplotlib.pyplot as plt
      fig = plt.figure()
      N = 400
      leg_handles = []
      leg_labels = []
      if not boundary:
        boundary = [min(self.X), max(self.X)]
      grid = np.linspace(boundary[0], boundary[1], N).reshape((N, 1))
      (pred_vals, pred_stds) = self.eval(grid, 'std')
      # Shade a high confidence region
      conf_band_up = pred_vals + conf_width * pred_stds
      conf_band_down = pred_vals - conf_width * pred_stds
      leg_conf = plt.fill_between(grid.ravel(), conf_band_up, conf_band_down,
                                  color=[0.9, 0.9, 0.9])
      # Plot the samples
      gp_samples = self.draw_samples(num_samples, grid)
      plt.plot(grid, gp_samples.T, '--', linewidth=0.5)
      # plot the true function if available.
      if true_func:
        leg_true = plt.plot(grid, true_func(grid), 'b--', linewidth=3,
                            label='true function')
        leg_handles.append(leg_true)
      # Plot the posterior mean
      leg_post_mean = plt.plot(grid, pred_vals, 'k-', linewidth=4,
                               label='post mean')
      # Finally plot the training data.
      leg_data = plt.plot(self.X, self.Y, 'kx', mew=4, markersize=10,
                          label='data')
      # Finally either plot or show the figure
      if file_name is None:
        plt.show()
      else:
        fig.savefig(file_name)


class GPFitter(object):
  """ Class for fitting Gaussian processes. """
  # pylint: disable=attribute-defined-outside-init
  # pylint: disable=no-value-for-parameter

  def __init__(self, X, Y, options, reporter='default'):
    """ Constructor. """
    super(GPFitter, self).__init__()
    assert len(X) == len(Y)
    self.reporter = get_reporter(reporter)
    if isinstance(options, list):
      options = load_options(options, 'GP', reporter=self.reporter)
    self.options = options
    self.X = X
    self.Y = Y
    self.num_data = len(X)
    self._set_up()

  def _set_up(self):
    """ Sets up a bunch of ancillary parameters. """
    # Set place-holders for hp_bounds or hp_vals depending on whether we are using ml or
    # post_sampling/mean
    # -----------------------------------------------------------------------------
    self.cts_hp_bounds = [] # The bounds for each hyper parameter should be a
                            # num_cts_hps x 2 array where the 1st/2nd columns are the
                            # lowe/upper bounds.
    self.dscr_hp_vals = []  # A list of lists where each list contains a list of values
                            # for each discrete hyper-parameter.
    self.param_order = []
    # Set up for common things - including mean and noise variance hyper-parameters
    # -----------------------------------------------------------------------------
    epsilon = 0.0001
    self.Y_var = np.array(self.Y).std() ** 2 + epsilon if len(self.Y) > 0 else \
                 epsilon
    self._set_up_mean_and_noise_variance_bounds()
    # Set up hyper-parameters for the child.
    # -----------------------------------------------------------------------------
    self._child_set_up()
    self._hp_tune_method_set_up()
    self.cts_hp_bounds = np.array(self.cts_hp_bounds)
    # Some post child set up
    if 'ml' in self.methods_to_use:
      # The number of hyper parameters
      self.num_hps = len(self.cts_hp_bounds) + len(self.dscr_hp_vals)
      self._set_up_ml_hp_tune()
    if 'post_sampling' in self.methods_to_use:
      self._set_up_post_sampling_hp_tune()
      self.num_hps = len(self.hp_priors) # The number of hyper parameters
    if 'post_mean' in self.methods_to_use:
      self._set_up_post_mean_hp_tune()
      self.num_hps = len(self.hp_priors) # The number of hyper parameters

  def _hp_tune_method_set_up(self):
    """ Sets up probabilities for hp tune methods. """
    self.methods_to_use = [elem.lower() for elem in
                           self.options.hp_tune_criterion.split('-')]
    for method in self.methods_to_use:
      if method not in ['ml', 'post_sampling', 'post_mean']:
        raise ValueError('hp_tune_criterion should be ml or post_sampling.')
    self.methods_to_use_counter = {key: 0 for key in self.methods_to_use}
    if self.options.hp_tune_probs == 'uniform':
      self.hp_tune_probs = np.ones(len(self.methods_to_use)) /\
                           float(len(self.methods_to_use))
    elif self.options.hp_tune_probs == 'adaptive':
      self.hp_tune_uniform_sampling_prob = 0.05
      self.hp_tune_sampling_weights = {key: 1.0 for key in self.methods_to_use}
      self.hp_tune_probs = self._get_adaptive_hp_tune_probs()
    else:
      self.hp_tune_probs = np.array([float(x) for x in
                                     self.options.hp_tune_probs.split('-')])
      if len(self.hp_tune_probs) != len(self.methods_to_use):
        self.hp_tune_probs = np.ones(len(self.methods_to_use)) /\
                             float(len(self.methods_to_use))
    self.hp_tune_probs = self.hp_tune_probs / self.hp_tune_probs.sum()

  def _get_adaptive_hp_tune_probs(self):
    """ Computes the adaptive hyper parameter tuning method probs. """
    num_methods = len(self.methods_to_use)
    uniform_sampling_probs = self.hp_tune_uniform_sampling_prob * \
                             np.ones((num_methods,)) / num_methods
    hp_tune_succ_counter = np.array([self.hp_tune_sampling_weights[key] for
                                     key in self.methods_to_use])
    hp_tune_use_counter = np.array([self.methods_to_use_counter[key] for
                                    key in self.methods_to_use])
    hp_tune_weights = hp_tune_succ_counter / np.sqrt(1 + hp_tune_use_counter)
    hp_tune_norm_weights = hp_tune_weights / hp_tune_weights.sum()
    adaptive_sampling_probs = (1 - self.hp_tune_uniform_sampling_prob) * \
                              hp_tune_norm_weights
    ret = uniform_sampling_probs + adaptive_sampling_probs
    return ret / ret.sum()

  def _set_up_mean_and_noise_variance_bounds(self):
    """ Sets up bounds for the mean value and the noise. """
    if not (hasattr(self.options, 'mean_func') and self.options.mean_func is not None) \
      and self.options.mean_func_type == 'tune':
      Y_std = np.sqrt(self.Y_var)
      if len(self.Y) > 0:
        Y_median = np.median(self.Y)
        Y_half_range = 0.5 * (max(self.Y) - min(self.Y)) if len(self.Y) > 0 else 1.0
      else:
        Y_median = 0.0
        Y_half_range = 1.0
      Y_width = 0.5 * (Y_half_range + Y_std)
      self.mean_func_bounds = [Y_median - 3 * Y_width, Y_median + 3 * Y_width]
      self.cts_hp_bounds.append(self.mean_func_bounds)
      self.param_order.append(["noise_mean", "cts"])
    # 2. The noise variance
    if self.options.noise_var_type == 'tune':
      self.noise_var_log_bounds = [np.log(0.005 * self.Y_var), np.log(0.2 * self.Y_var)]
      self.cts_hp_bounds.append(self.noise_var_log_bounds)
      self.param_order.append(["noise_var", "cts"])

  def _child_set_up(self):
    """ Here you should set up parameters for the child, such as the bounds for the
        optimiser etc. """
    raise NotImplementedError('Implement _child_set_up in a child method.')

  def _set_up_ml_hp_tune(self):
    """ Sets up optimiser for ml hp tune. """
    # pylint: disable=too-many-branches
    # define the following internal functions to abstract things out more.
    def _direct_wrap(*args):
      """ A wrapper so as to only return the optimal point. """
      opt_val, opt_pt, _ = direct_ft_maximise(*args)
      return opt_val, opt_pt, None
    def _pdoo_wrap(*args):
      """ A wrapper so as to only return the optimal point. """
      opt_val, opt_pt, _ = pdoo_maximise(*args)
      return opt_val, opt_pt, None
    def _rand_wrap(*args):
      """ A wrapper so as to only return the optimal point. """
      opt_val, opt_pt, _ = random_maximise(*args, vectorised=False)
      return opt_val, opt_pt, None
    def _rand_exp_sampling_wrap(*args):
      """ A wrapper so as to only return the optimal point. """
      sample_cts_hps, sample_dscr_hps, lml_vals = \
        random_sample_cts_dscr(*args, vectorised=False)
      sample_probs = np.exp(lml_vals - max(lml_vals))
      sample_probs = sample_probs / sample_probs.sum()
      return sample_cts_hps, sample_dscr_hps, sample_probs
    # Set some parameters
    if self.options.ml_hp_tune_opt == 'default':
      self.ml_hp_tune_opt_method = _get_default_ml_hp_tune_opt(self.num_hps)
    else:
      self.ml_hp_tune_opt_method = self.options.ml_hp_tune_opt
    if (hasattr(self.options, 'hp_tune_max_evals') and
        self.options.hp_tune_max_evals is not None and
        self.options.hp_tune_max_evals > 0):
      self.hp_tune_max_evals = self.options.hp_tune_max_evals
    else:
      if self.ml_hp_tune_opt_method in ['direct', 'pdoo']:
        self.hp_tune_max_evals = min(1e4, max(500, self.num_hps * 50))
      elif self.ml_hp_tune_opt_method == 'rand':
        self.hp_tune_max_evals = min(1e4, max(500, self.num_hps * 200))
      elif self.ml_hp_tune_opt_method == 'rand_exp_sampling':
        self.hp_tune_max_evals = min(1e5, max(500, self.num_hps * 400))
    # Now set up the function that will be used for optimising
    if self.ml_hp_tune_opt_method == 'direct':
      self.cts_hp_optimise = lambda obj, max_evals: _direct_wrap(obj, self.cts_hp_bounds,
                                                                 max_evals)
    elif self.ml_hp_tune_opt_method == 'pdoo':
      self.cts_hp_optimise = lambda obj, max_evals: _pdoo_wrap(obj, self.cts_hp_bounds,
                                                               max_evals)
    elif self.ml_hp_tune_opt_method == 'rand':
      self.cts_hp_optimise = lambda obj, max_evals: _rand_wrap(obj, self.cts_hp_bounds,
                                                               max_evals)
    elif self.ml_hp_tune_opt_method == 'rand_exp_sampling':
      self.hp_sampler = lambda obj, max_evals: _rand_exp_sampling_wrap(obj, \
                          self.cts_hp_bounds, self.dscr_hp_vals, max_evals)

  def _set_up_post_sampling_hp_tune(self):
    """ Sets up posterior sampling for tuning the parameters of the GP. """
    if self.options.post_hp_tune_method == 'slice':
      self.hp_sampler_cts = lambda model, init_sample, num_samples, burn: \
                               model.draw_samples('slice', num_samples, init_sample, burn)
    elif self.options.post_hp_tune_method == 'nuts':
      self.hp_sampler_cts = lambda model, init_sample, num_samples, burn: \
                                model.draw_samples('nuts', num_samples, init_sample, burn)

    self.hp_sampler_dscr = lambda model, init_sample, num_samples: \
                                model.draw_samples('metropolis', num_samples, init_sample)

    # Priors for all the hyper parameters
    self.hp_priors = []
    for _, bounds in enumerate(self.cts_hp_bounds):
      self.hp_priors.append(continuous.ContinuousUniform(bounds[0], bounds[-1]))

    for _, vals in enumerate(self.dscr_hp_vals):
      self.hp_priors.append(discrete.Categorical(vals, np.repeat(1.0/len(vals),
                                                                 len(vals))))

  def _set_up_post_mean_hp_tune(self):
    """ Sets up using the posterior mean for tuning the parameters of the GP. """
    raise NotImplementedError('Not implemented post_mean yet.')

  def build_gp(self, gp_cts_hps, gp_dscr_hps, other_gp_params=None, *args, **kwargs):
    """ A method which builds a GP from the given gp_hyperparameters. It calls
        _child_build_gp after running some checks. """
    # pylint: disable=too-many-branches
    # Check the length of the hyper-parameters
    if self.num_hps != len(gp_cts_hps) + len(gp_dscr_hps):
      raise ValueError('gp_hyperparams should be of length %d. Given length: %d.'%(
          self.num_hps, len(gp_cts_hps) + len(gp_dscr_hps)))
    # Compute the mean and the noise variance first
    # Mean function -------------------------------------
    if hasattr(self.options, 'mean_func') and self.options.mean_func is not None:
      mean_func = self.options.mean_func
    else:
      if self.options.mean_func_type == 'mean':
        mean_func_const_value = np.mean(self.Y)
      elif self.options.mean_func_type == 'median':
        mean_func_const_value = np.median(self.Y)
      elif self.options.mean_func_type == 'upper_bound':
        mean_func_const_value = np.mean(self.Y) + 3 * np.std(self.Y)
      elif self.options.mean_func_type == 'const':
        mean_func_const_value = self.options.mean_func_const
      elif self.options.mean_func_type == 'tune':
        mean_func_const_value = np.asscalar(gp_cts_hps[0])
        gp_cts_hps = gp_cts_hps[1:]
      else:
        mean_func_const_value = 0
      def _get_mean_func(_mean_func_const_value):
        """ Returns the mean function from the constant value. """
        return lambda x: np.array([_mean_func_const_value] * len(x))
      mean_func = _get_mean_func(mean_func_const_value)
    # Noise variance ------------------------------------
    if self.options.noise_var_type == 'tune':
      noise_var = np.exp(gp_cts_hps[0])
      gp_cts_hps = gp_cts_hps[1:]
    elif self.options.noise_var_type == 'label':
      noise_var = self.options.noise_var_label * (self.Y.std() ** 2)
    else:
      noise_var = self.options.noise_var_value
    ret_gp, ret_cts_hps, ret_dscr_hps = self._child_build_gp(mean_func, noise_var, \
       gp_cts_hps, gp_dscr_hps, other_gp_params=other_gp_params, *args, **kwargs)
    assert len(ret_cts_hps) == 0
    assert len(ret_dscr_hps) == 0
    return ret_gp

  def _child_build_gp(self, mean_func, noise_var, gp_cts_hps, gp_dscr_hps,
                      other_gp_params=None, *args, **kwargs):
    """ A method which builds the child GP from the given gp_hyperparameters. Should be
        implemented in a child method. """
    raise NotImplementedError('Implement _child_build_gp in a child method.')

  def _tuning_objective(self, gp_cts_hps, gp_dscr_hps, other_gp_params=None,
                        *args, **kwargs):
    """ This function computes the tuning objective (such as the marginal likelihood)
        which is to be maximised in fit_gp. """
    built_gp = self.build_gp(gp_cts_hps, gp_dscr_hps, other_gp_params=other_gp_params,
                             *args, **kwargs)
    if 'ml' in self.methods_to_use or 'marginal_likelihood' in self.methods_to_use:
      ret = built_gp.compute_log_marginal_likelihood()
    elif 'cv' in self.methods_to_use or 'cross_validation' in self.methods_to_use:
      raise NotImplementedError('Yet to implement cross validation based hp-tuning.')
    else:
      raise ValueError('hp_tune_criterion should be either ml or cv')
    return ret

  def _tuning_objective_post_sampling(self, gp_cts_hps, gp_dscr_hps, \
       other_gp_params=None, param=None, param_num=None, gradient=False, *args, **kwargs):
    """ This function computes the tuning objective for posterior sampling case. """
    built_gp = self.build_gp(gp_cts_hps, gp_dscr_hps, other_gp_params=other_gp_params,
                             *args, **kwargs)
    if gradient:
      ret = built_gp.compute_grad_log_marginal_likelihood(param, param_num)
    else:
      ret = built_gp.compute_log_marginal_likelihood()
    return ret

  def _optimise_cts_hps_for_given_dscr_hps(self, given_dscr_hps):
    """ Optimises the continuous hyper-parameters for the given discrete hyper-params.
        Can be overridden by a child class if necessary.
    """
    cts_tuning_obj = lambda arg: self._tuning_objective(arg, list(given_dscr_hps))
    opt_cts_val, opt_cts_hps, _ = self.cts_hp_optimise(cts_tuning_obj,
                                                       self.hp_tune_max_evals)
    return opt_cts_val, opt_cts_hps, None

  def _sample_cts_dscr_hps_for_rand_exp_sampling(self):
    """ Samples continous and discrete hyper-parameters for rand_exp_sampling. """
    sample_cts_hps, sample_dscr_hps, sample_probs = \
      self.hp_sampler(self._tuning_objective, self.hp_tune_max_evals)
    sample_other_gp_params = [None] * len(sample_cts_hps)
    return sample_cts_hps, sample_dscr_hps, sample_other_gp_params, sample_probs

  def _sample_cts_dscr_hps_for_post_sampling(self, num_samples):
    """ Samples continous and discrete hyper-parameters for post_sampling. """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    def _logp(x):
      """ Computes log probability of observations at x. """
      # sum of log probability priors
      if self.parameter == 'additive_grp':
        if x < 0:
          return -np.inf
        permut = list(np.random.RandomState(seed=x[0]).permutation(self.add_dim))
        groupings = [permut[i:i+self.group_size]
                     for i in range(0, self.add_dim, self.group_size)]
        self.other_gp_params = Namespace(add_gp_groupings=groupings)
      elif type(self.hp_priors[self.curr_hp]).__name__ == "Categorical":
        self.hps[self.curr_hp] = self.hp_priors[self.curr_hp].get_category(np.asscalar(x))
      else:
        self.hps[self.curr_hp] = x

      lp = 0
      for i, prior in enumerate(self.hp_priors):
        if type(self.hp_priors[i]).__name__ == "Categorical":
          lp += prior.logp(self.hp_priors[i].get_id(self.hps[i]))
        else:
          lp += prior.logp(self.hps[i])
      if not np.isfinite(lp):
        return lp

      lp += self._tuning_objective_post_sampling(self.hps[0:len(self.cts_hp_bounds)], \
                self.hps[len(self.cts_hp_bounds):self.num_hps], self.other_gp_params)
      return lp

    def _pdf(x):
      """ Computes probability of observations at x. """
      val = _logp(x)
      if not np.isfinite(val):
        return val
      return np.exp(val)

    def _grad_logp(x):
      """ Computes gradient of log probability of observations at x. """
      if not np.isfinite(self.hp_priors[self.curr_hp].logp(x)):
        return 0
      self.hps[self.curr_hp] = x
      lp = self.hp_priors[self.curr_hp].grad_logp(self.hps[self.curr_hp])
      lp += self._tuning_objective_post_sampling(self.hps[0:len(self.cts_hp_bounds)], \
                self.hps[len(self.cts_hp_bounds):self.num_hps], self.other_gp_params, \
                self.parameter, self.param_num, True)
      return lp

    offset = self.options.post_hp_tune_offset
    _model = Model(_pdf, _logp, _grad_logp)
    self.hps = np.ones((self.num_hps,))
    best_cts_hps = np.zeros([num_samples, len(self.cts_hp_bounds)])
    best_dscr_hps = np.zeros([num_samples, len(self.dscr_hp_vals)])
    best_other_gp_params = [None] * num_samples
    total_samples = (num_samples - 1)*offset + 1
    cts_hps = np.zeros([total_samples, len(self.cts_hp_bounds)])
    dscr_hps = np.zeros([total_samples, len(self.dscr_hp_vals)])
    _other_gp_params = [None] * total_samples
    self.param_num = -1

    # burn value
    if self.options.post_hp_tune_burn == -1:
      burn = int(np.sqrt(self.num_hps)*100)
      np.clip(burn, 100, 2000)
    else:
      burn = self.options.post_hp_tune_burn

    # Initializing hyper parameters
    for i in range(len(self.cts_hp_bounds)):
      self.hps[i] = self.hp_priors[i].get_mean()
    for i in range(len(self.cts_hp_bounds), self.num_hps):
      self.hps[i] = self.hp_priors[i].draw_samples('random', 1)
      if type(self.hp_priors[i]).__name__ == "Categorical":
        self.hps[i] = self.hp_priors[i].get_category(int(self.hps[i]))

    # Grouping intitialization for additive gp
    if vars(self.options).get('use_additive_gp', False) or \
       vars(self.options).get('domain_use_additive_gp', False):
      if vars(self.options).get('domain_use_additive_gp', False):
        self.add_dim = self.domain_dim
        add_max_group_size = self.domain_add_max_group_size
      else:
        self.add_dim = self.dim
        add_max_group_size = self.add_max_group_size
      permut = list(np.random.permutation(self.add_dim))
      groupings = [permut[i:i+int(self.hps[-1])]
                   for i in range(0, self.add_dim, int(self.hps[-1]))]
      self.other_gp_params = Namespace(add_gp_groupings=groupings)
    else:
      self.other_gp_params = Namespace(add_gp_groupings=None)

    # Randomly selcting a parameter and sampling it using tuning objective
    order = list(range(self.num_hps))
    np.random.shuffle(order)
    for i in order:
      self.curr_hp = i
      self.parameter = self.param_order[i][0]
      if self.parameter == "dim_bandwidths":
        self.param_num = self.param_num + 1
      dscr_i = i - len(self.cts_hp_bounds)
      if self.param_order[i][-1] == "cts":
        cts_hps[:, i] = np.squeeze(self.hp_sampler_cts(_model, self.hps[i],
                                                       total_samples, burn), axis=1)
        self.hps[i] = cts_hps[0, i]
      elif self.parameter != "additive_grp":
        if type(self.hp_priors[i]).__name__ == "Categorical":
          init_sample = self.hp_priors[i].get_id(self.hps[i])
          dscr_hp = np.squeeze(self.hp_sampler_dscr(_model, init_sample, total_samples),
                               axis=1)
          for j, val in enumerate(dscr_hp):
            dscr_hps[j, dscr_i] = self.hp_priors[i].get_category(int(val))
        else:
          dscr_hps[:, dscr_i] = np.squeeze(self.hp_sampler_dscr(_model, \
                                                self.hps[i], total_samples), axis=1)
        self.hps[i] = dscr_hps[0, dscr_i]
      else:
        dscr_hps[:, dscr_i] = np.random.randint(1, add_max_group_size + 1, total_samples)
        seed = int(np.random.randint(self.add_dim))
        for j in range(total_samples):
          self.group_size = int(dscr_hps[j, dscr_i])
          seed = int(np.squeeze(self.hp_sampler_dscr(_model, seed, 1), axis=1))
          permut = list(np.random.RandomState(seed=seed).permutation(self.add_dim))
          groupings = [permut[k:k+self.group_size]
                       for k in range(0, self.add_dim, self.group_size)]
          _other_gp_params[j] = Namespace(add_gp_groupings=groupings)
        self.other_gp_params = _other_gp_params[0]

    for i in range(num_samples):
      best_cts_hps[i, :] = cts_hps[i*offset, :]
      best_dscr_hps[i, :] = dscr_hps[i*offset, :]
      best_other_gp_params[i] = _other_gp_params[i*offset]

    return best_cts_hps, best_dscr_hps, best_other_gp_params

  def get_next_gp(self):
    """ Gets the GP to use in the current iteration. """
    if self.options.hp_tune_probs == 'adaptive':
      self.hp_tune_probs = self._get_adaptive_hp_tune_probs()
    method = np.random.choice(self.methods_to_use, p=self.hp_tune_probs)
    fit_type = self.hp_tune_results[method][0]
    if fit_type in ['fitted_gp', 'post_fitted_gp']:
      gp = self.hp_tune_results[method][1]
    elif fit_type in ['sample_hps_with_probs', 'post_sample_hps_with_probs']:
      next_gp_hps = self.hp_tune_results[method][1].pop(0)
      self.hp_tune_results[method][1].append(next_gp_hps)
      gp = self.build_gp(next_gp_hps[0], next_gp_hps[1], other_gp_params=next_gp_hps[2],
                         build_posterior=False)
    return (fit_type, method, gp)

  def update_hp_tune_method_weight(self, method, weight_to_add=1):
    """ Updates hp_tune method weight in the case of adaptive tuning. """
    if self.options.hp_tune_probs == 'adaptive':
      self.hp_tune_sampling_weights[method] += weight_to_add

  def fit_gp_for_gp_bandit(self, num_samples=1):
    """ Fits a GP according to the tuning criterions. """
    self.hp_tune_results = {}
    for method in self.methods_to_use:
      ret = self.fit_gp(num_samples, method)
      if ret[0] == 'fitted_gp':
        self.hp_tune_results[method] = (ret[0], ret[1])
      elif ret[0] == 'post_fitted_gp':
        self.hp_tune_results[method] = (ret[0], ret[1])
      elif ret[0] == 'sample_hps_with_probs':
        # in this case, ret is a 5-tuple where ret[0] is sample_hps_with_probs, ret[1] is
        # a list of continuous hps, ret[2] is a list of discrete hps, ret[3] is either
        # None or a namespace containing the field add_gp_groupings.
        # See gp.gp_core.fit_gp
        sample_hps = list(zip(ret[1], ret[2], ret[3]))
        sample_probs = ret[-1]
        # Sample the hyper_parameters we will use until the next time we build a model.
        if sum(sample_probs > 0) >= num_samples:
          to_replace = self.options.rand_exp_sampling_replace
        else:
          to_replace = True
        use_hps_idxs = np.random.choice(len(sample_hps),
                                        size=(num_samples,),
                                        replace=to_replace,
                                        p=sample_probs)
        self.hp_tune_results[method] = (ret[0], [sample_hps[idx] for idx in use_hps_idxs])
      elif ret[0] == 'post_sample_hps_with_probs':
        # in this case, ret is a 4-tuple where ret[0] is post_sample_hps_with_probs,
        # ret[1] is a list of continuous hps, ret[2] is a list of discrete hps, ret[3]
        # is either None or a namespace containing the field add_gp_groupings.
        # See gp.gp_core.fit_gp
        self.hp_tune_results[method] = (ret[0], list(zip(ret[1], ret[2], ret[3])))
      else:
        raise ValueError('Unknown option %s for results of fit_gp.'%(ret[0]))

  def fit_gp(self, num_samples=1, hp_tune_criterion=None):
    """ Fits a GP according to the tuning criterion. Returns the best GP along with the
        hyper-parameters. """
    if hp_tune_criterion is None:
      hp_tune_criterion = self.options.hp_tune_criterion
    if hp_tune_criterion == 'ml':
      if self.ml_hp_tune_opt_method in ['direct', 'rand', 'pdoo']:
        best_cts_hps = None
        best_dscr_hps = None
        best_other_params = None
        best_hps_val = -np.inf
        for dscr_hps in itertools_product(*self.dscr_hp_vals):
          opt_cts_val, opt_cts_hps, opt_other_params = \
             self._optimise_cts_hps_for_given_dscr_hps(dscr_hps)
          if opt_cts_val > best_hps_val:
            best_cts_hps = list(opt_cts_hps)
            best_dscr_hps = list(dscr_hps)
            best_other_params = opt_other_params
            best_hps_val = opt_cts_val
        opt_gp = self.build_gp(best_cts_hps, best_dscr_hps,
                               other_gp_params=best_other_params)
        opt_hps = (best_cts_hps, best_dscr_hps)
        return 'fitted_gp', opt_gp, opt_hps
      elif self.ml_hp_tune_opt_method == 'rand_exp_sampling':
        sample_cts_hps, sample_dscr_hps, sample_other_gp_params, sample_probs = \
          self._sample_cts_dscr_hps_for_rand_exp_sampling()
        return ('sample_hps_with_probs', sample_cts_hps, sample_dscr_hps,
                sample_other_gp_params, sample_probs)
    elif hp_tune_criterion == 'post_sampling':
      sample_cts_hps, sample_dscr_hps, sample_other_gp_params = \
        self._sample_cts_dscr_hps_for_post_sampling(num_samples)
      if num_samples == 1:
        opt_gp = self.build_gp(sample_cts_hps[0], sample_dscr_hps[0],
                               other_gp_params=sample_other_gp_params[0])
        opt_hps = (sample_cts_hps, sample_dscr_hps)
        return 'post_fitted_gp', opt_gp, opt_hps
      else:
        return ('post_sample_hps_with_probs', sample_cts_hps, sample_dscr_hps,
                sample_other_gp_params)

      # GPFitter class ends here ---------------------------------------------------------


# Some utilities we will be using above -------------------------------------------------
def _get_cholesky_decomp(K_trtr_wo_noise, noise_var, handle_non_psd_kernels):
  """ Computes cholesky decomposition after checking how to handle non-psd kernels. """
  if handle_non_psd_kernels == 'try_before_project':
    K_trtr_w_noise = K_trtr_wo_noise + noise_var * np.eye(K_trtr_wo_noise.shape[0])
    try:
      # If the cholesky decomposition on the (noise added) matrix succeeds, return!
      L = stable_cholesky(K_trtr_w_noise, add_to_diag_till_psd=False)
      return L
    except np.linalg.linalg.LinAlgError:
      # otherwise, project and return
      return _get_cholesky_decomp(K_trtr_wo_noise, noise_var, 'project_first')
  elif handle_non_psd_kernels == 'project_first':
    # project the Kernel (without noise) to the PSD cone and return
    K_trtr_wo_noise = project_symmetric_to_psd_cone(K_trtr_wo_noise)
    return _get_cholesky_decomp(K_trtr_wo_noise, noise_var, 'guaranteed_psd')
  elif handle_non_psd_kernels == 'guaranteed_psd':
    K_trtr_w_noise = K_trtr_wo_noise + noise_var * np.eye(K_trtr_wo_noise.shape[0])
    return stable_cholesky(K_trtr_w_noise)
  else:
    raise ValueError('Unknown option for handle_non_psd_kernels: %s'%(
        handle_non_psd_kernels))

def get_post_covar_from_raw_covar(raw_post_covar, noise_var, is_guaranteed_psd):
  """ Computes the posterior covariance from the raw_post_covar. This is mostly to
      account for the fact that the kernel may not be psd.
  """
  if is_guaranteed_psd:
    return raw_post_covar
  else:
    epsilon = 0.05 * noise_var
    return project_symmetric_to_psd_cone(raw_post_covar, epsilon=epsilon)




"""
  Metropolis Hastings Sampler -- Adapted from pymc3 library.
  https://github.com/pymc-devs/pymc3
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=unexpected-keyword-arg
# pylint: disable=redefined-variable-type

import scipy.linalg
import numpy as np
import numpy.random as nr

__all__ = ['Metropolis', 'BinaryMetropolis', 'NormalProposal', 'CauchyProposal',
           'LaplaceProposal', 'PoissonProposal', 'MultivariateNormalProposal']

# Available proposal distributions for Metropolis

class Proposal(object):
  """ Base class for Proposal Distribution."""
  def __init__(self, s):
    self.s = s


class NormalProposal(Proposal):
  """ Normal Proposal Distribution."""
  def __call__(self):
    return nr.normal(scale=self.s)


class UniformProposal(Proposal):
  """ Uniform Proposal Distribution."""
  def __call__(self):
    return nr.uniform(low=-self.s, high=self.s, size=len(self.s))


class CauchyProposal(Proposal):
  """ Cauchy Proposal Distribution."""
  def __call__(self):
    return nr.standard_cauchy(size=np.size(self.s)) * self.s


class LaplaceProposal(Proposal):
  """ Laplace Proposal Distribution."""
  def __call__(self):
    size = np.size(self.s)
    return (nr.standard_exponential(size=size) - nr.standard_exponential(size=size))\
            *self.s


class PoissonProposal(Proposal):
  """ Poisson Proposal Distribution."""
  def __call__(self):
    return nr.poisson(lam=self.s, size=np.size(self.s)) - self.s


class MultivariateNormalProposal(Proposal):
  """ Multivariate Normal Proposal Distribution."""
  def __init__(self, s):
    n, m = s.shape
    if n != m:
      raise ValueError("Covariance matrix is not symmetric.")
    self.n = n
    self.chol = scipy.linalg.cholesky(s, lower=True)
    super(MultivariateNormalProposal, self).__init__()

  def __call__(self, num_draws=None):
    if num_draws is not None:
      b = np.random.randn(self.n, num_draws)
      return np.dot(self.chol, b).T
    else:
      b = np.random.randn(self.n)
      return np.dot(self.chol, b)


class Metropolis(object):
  """
  Metropolis-Hastings sampling step

  Parameters
  ----------
  S : standard deviation or covariance matrix
    Some measure of variance to parameterize proposal distribution
  proposal_dist : function
    Function that returns zero-mean deviates when parameterized with
    S (and n). Defaults to normal.
  scaling : scalar or array
    Initial scale factor for proposal. Defaults to 1.
  tune : bool
    Flag for tuning. Defaults to True.
  tune_interval : int
    The frequency of tuning. Defaults to 100 iterations.
  """

  def __init__(self, model, discrete=False, S=None, proposal_dist=None, scaling=1.,\
               tune=True, tune_interval=100):

    if S is None:
      S = np.ones(model.get_dim())

    if proposal_dist is not None:
      self.proposal_dist = proposal_dist(S)
    elif S.ndim == 1:
      self.proposal_dist = NormalProposal(S)
    elif S.ndim == 2:
      self.proposal_dist = MultivariateNormalProposal(S)
    else:
      raise ValueError("Invalid rank for variance: %s" % S.ndim)

    self.model = model
    self.scaling = np.atleast_1d(scaling).astype('d')
    self.tune = tune
    self.tune_interval = tune_interval
    self.steps_until_tune = tune_interval
    self.accepted = 0
    self.discrete = discrete

    super(Metropolis, self).__init__()

  def _sample(self, q0):
    """
    Helper function which implements metropolis sampler.
    """
    if not self.steps_until_tune and self.tune:
      # Tune scaling parameter
      self.scaling = tune_params(self.scaling, self.accepted / float(self.tune_interval))
      # Reset counter
      self.steps_until_tune = self.tune_interval
      self.accepted = 0

    delta = self.proposal_dist() * self.scaling
    if self.discrete:
      delta = np.round(delta, 0).astype('int64')
      q0 = q0.astype('int64')
      q = (q0 + delta).astype('int64')
    else:
      q = q0 + delta

    accept = delta_logp(self.model, q, q0)
    q_new, accepted = metrop_select(accept, q, q0)
    self.accepted += accepted

    self.steps_until_tune -= 1

    return q_new

  def sample(self, q0, num_samples=1):
    """
    sample -- function which populates samples
    by calling function _sample iteratively
    """
    if num_samples is None:
      num_samples = 1
    if not hasattr(q0, '__len__'):
      q0 = np.array([q0])

    samples = np.zeros([num_samples, len(q0)])
    for i in range(num_samples):
      samples[i] = self._sample(q0)
      q0 = samples[i]

    return samples


class BinaryMetropolis(object):
  """Metropolis-Hastings optimized for binary variables

  Parameters
  ----------
  scaling : scalar or array
      Initial scale factor for proposal. Defaults to 1.
  """

  def __init__(self, model, scaling=1.):
    self.model = model
    self.scaling = scaling

    super(BinaryMetropolis, self).__init__()

  def _sample(self, q0):
    """
    Helper function which implements binary metropolis sampler.
    """
    # Convert adaptive_scale_factor to a jump probability
    p_jump = 1. - .5 ** self.scaling

    rand_array = nr.random(q0.shape)
    q = np.copy(q0)

    # Locations where switches occur, according to p_jump
    switch_locs = (rand_array < p_jump)
    q[switch_locs] = True - q[switch_locs]

    accept = self.model.logp(q) - self.model.logp(q0)
    q_new, _ = metrop_select(np.exp(accept), q, q0)

    return q_new

  def sample(self, q0, num_samples=1):
    """
    sample -- function which populates samples
    by calling function _sample iteratively
    """
    if num_samples is None:
      num_samples = 1
    if not hasattr(q0, '__len__'):
      q0 = np.array([q0])

    samples = np.zeros([num_samples, len(q0)])
    for i in range(num_samples):
      samples[i] = self._sample(q0)
      q0 = samples[i]

    return samples


def metrop_select(mr, q, q0):
  """Perform rejection/acceptance step for Metropolis class samplers.

  Returns the new sample q if a uniform random number is less than the
  metropolis acceptance rate (`mr`), and the old sample otherwise, along
  with a boolean indicating whether the sample was accepted.

  Parameters
  ----------
  mr : float, Metropolis acceptance rate
  q : proposed sample
  q0 : current sample

  Returns
  -------
  q or q0
  """
  # Compare acceptance ratio to uniform random number
  if np.isfinite(mr) and nr.uniform() < mr:
    return q, True
  else:
    return q0, False


def tune_params(scale, acc_rate):
  """
  Tunes the scaling parameter for the proposal distribution
  according to the acceptance rate over the last tune_interval:

  Rate  Variance adaptation
  ----  -------------------
  <0.001    x 0.1
  <0.05     x 0.5
  <0.2      x 0.9
  >0.5      x 1.1
  >0.75     x 2
  >0.95     x 10

  """

  # Switch statement
  if acc_rate < 0.001:
    # reduce by 90 percent
    scale *= 0.1
  elif acc_rate < 0.05:
    # reduce by 50 percent
    scale *= 0.5
  elif acc_rate < 0.2:
    # reduce by ten percent
    scale *= 0.9
  elif acc_rate > 0.95:
    # increase by factor of ten
    scale *= 10.0
  elif acc_rate > 0.75:
    # increase by double
    scale *= 2.0
  elif acc_rate > 0.5:
    # increase by ten percent
    scale *= 1.1

  return scale


def delta_logp(model, q, q0):
  """ Function for calculating acceptance criteria """
  val = model.logp(q) - model.logp(q0)
  val = min(1, np.exp(val))

  return val



"""
  Implements Genetic algorithms for black-box optimisation.
  --kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-member

from argparse import Namespace
from numpy.random import choice
# Local imports
from .blackbox_optimiser import BlackboxOptimiser, blackbox_opt_args
from ..utils.general_utils import sample_according_to_exp_probs
from ..utils.option_handler import get_option_specs, load_options
from ..utils.reporters import get_reporter

ga_specific_opt_args = [
  get_option_specs('num_mutations_per_epoch', False, 5,
    'Number of mutations per epoch.'),
  get_option_specs('num_candidates_to_mutate_from', False, -1,
    'The number of candidates to choose the mutations from.'),
  get_option_specs('fitness_sampler_scaling_const', False, 2,
    'The scaling constant for sampling according to exp_probs.'),
  ]

ga_opt_args = ga_specific_opt_args + blackbox_opt_args


class GAOptimiser(BlackboxOptimiser):
  """ Class for optimisation based on Genetic algorithms. """

  def __init__(self, func_caller, worker_manager, mutation_op, crossover_op=None,
               options=None, reporter=None):
    """ Constructor.
      mutation_op: A function which takes in a list of objects and modifies them.
      crossover_op: A function which takes in two objects and performs a cross-over
                    operation.
      So far we have not implemented cross-over but included here in case we want to
      include it in the future.
      For other arguments, see BlackboxOptimiser
    """
    # TODO: implement cross-over operation
    if options is None:
      reporter = get_reporter(reporter)
      options = load_options(ga_opt_args, reporter=reporter)
    super(GAOptimiser, self).__init__(func_caller, worker_manager, model=None,
                                      options=options, reporter=reporter)
    self.mutation_op = mutation_op
    self.crossover_op = crossover_op
    self.to_eval_points = []

  def _opt_method_set_up(self):
    """ Additional set up. """
    # pylint: disable=attribute-defined-outside-init
    # Set up parameters for the mutations
    self.method_name = 'GA'
    self.num_mutations_per_epoch = self.options.num_mutations_per_epoch
    self.num_candidates_to_mutate_from = self.options.num_candidates_to_mutate_from

  def _opt_method_optimise_initialise(self):
    """ No initialisation for GA. """
    self.generate_new_eval_points()

  def _add_data_to_model(self, qinfos):
    """ Update the optimisation model. """
    pass

  def _child_build_new_model(self):
    """ Build new optimisation model. """
    pass

  def _get_candidates_to_mutate_from(self, num_mutations, num_candidates_to_mutate_from):
    """ Returns the candidates to mutate from. """
    all_prev_eval_points = self.prev_eval_points + self.history.query_points
    all_prev_eval_vals = self.prev_eval_vals + self.history.query_vals
    if num_candidates_to_mutate_from <= 0:
      idxs_to_mutate_from = sample_according_to_exp_probs(all_prev_eval_vals,
                              num_mutations, replace=True,
                              scaling_const=self.options.fitness_sampler_scaling_const,
                              sample_uniformly_if_fail=True)
      num_mutations_arg_to_mutation_op = [(idxs_to_mutate_from == i).sum() for i
                                          in range(len(all_prev_eval_points))]
      candidates_to_mutate_from = all_prev_eval_points
    else:
      cand_idxs_to_mutate_from = sample_according_to_exp_probs(all_prev_eval_vals,
        num_candidates_to_mutate_from, replace=False,
        scaling_const=self.options.fitness_sampler_scaling_const)
      candidates_to_mutate_from = [all_prev_eval_points[i] for i in
                                   cand_idxs_to_mutate_from]
      num_mutations_arg_to_mutation_op = num_mutations
    return candidates_to_mutate_from, num_mutations_arg_to_mutation_op

  def generate_new_eval_points(self, num_mutations=None,
                               num_candidates_to_mutate_from=None):
    """ Generates the mutations. """
    new_candidates = []
    num_tries = 0
    num_mutations_to_try = self.num_mutations_per_epoch if num_mutations is None \
                           else num_mutations
    while len(new_candidates) == 0:
      num_tries += 1
      generated_from_mutation_op = self.generate_new_eval_points_from_mutation_op(
                                     num_mutations_to_try, num_candidates_to_mutate_from)
      points_in_domain = [elem for elem in generated_from_mutation_op if
                          self.domain.is_a_member(elem)]
      new_candidates.extend(points_in_domain)
      if len(points_in_domain) == 0:
        if num_tries % 10 == 0:
          error_msg = ('Could not generate any points in domain from given mutation ' +
                       'operator despite %d tries with up to %d candidates.')%(num_tries,
                       num_mutations_to_try)
          self.reporter.writeln(error_msg)
        if num_tries == 35:
          error_msg = ('Could not generate any points in domain from given mutation ' +
            'operator despite %d tries with up to %d candidates. Quitting now.')%(
            num_tries, num_mutations_to_try)
          raise ValueError(error_msg)
        # Try a larger number of mutations the next time
        num_mutations_to_try = int(num_mutations_to_try * 1.2 + 1)
    new_candidates = new_candidates[:num_mutations]
    self.to_eval_points.extend(new_candidates)

  def generate_new_eval_points_from_mutation_op(self, num_mutations=None,
                                                num_candidates_to_mutate_from=None):
    """ Generates the mutations. """
    num_mutations = self.num_mutations_per_epoch if num_mutations is None else \
                      num_mutations
    num_candidates_to_mutate_from = self.num_candidates_to_mutate_from if \
      num_candidates_to_mutate_from is None else num_candidates_to_mutate_from
    candidates_to_mutate_from, num_mutations_arg_to_mutation_op = \
      self._get_candidates_to_mutate_from(num_mutations, num_candidates_to_mutate_from)
    new_eval_points = self.mutation_op(candidates_to_mutate_from,
                                       num_mutations_arg_to_mutation_op)
    return new_eval_points

  def _determine_next_query(self):
    """ Determine the next point for evaluation. """
    if len(self.to_eval_points) == 0:
      self.generate_new_eval_points()
    ret = self.to_eval_points.pop(0)
    return Namespace(point=ret)

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of eval points. Not implementing for now. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_method_str(self):
    """ Returns a string describing the method. """
    return 'ga'

  def is_an_mf_method(self):
    """ Returns False. """
    return False


# A GA optimiser with random fitness values ----------------------------------------------
class GARandOptimiser(GAOptimiser):
  """ Same as the GA optimiser, but the candidates to mutate from are picked randomly.
      This is used in the RAND baseline.
  """
  # pylint: disable=abstract-method

  def _child_set_up(self):
    """ Additional set up. """
    super(GARandOptimiser, self)._child_set_up()
    self.method_name = 'randGA'

  def _get_candidates_to_mutate_from(self, num_mutations, num_candidates_to_mutate_from):
    """ Returns a random list of points from the evaluations to mutate from. """
    all_prev_eval_points = self.prev_eval_points + self.history.query_points
    candidates_to_mutate_from = choice(all_prev_eval_points,
                                       self.num_candidates_to_mutate_from,
                                       replace=False)
    return candidates_to_mutate_from, num_mutations


# APIs
# ======================================================================================
def ga_optimise_from_args(func_caller, worker_manager, max_capital, mode, mutation_op,
                          is_rand=False, crossover_op=None, options=None,
                          reporter='default'):
  """ GA optimisation from args. """
  if options is None:
    reporter = get_reporter(reporter)
    options = load_options(ga_opt_args, reporter=reporter)
    options.mode = mode
  optimiser_class = GARandOptimiser if is_rand else GAOptimiser
  return (optimiser_class(func_caller, worker_manager, mutation_op, crossover_op,
                          options, reporter)).optimise(max_capital)




"""
  Unit tests for the class NNModifier in nn_modifiers.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from copy import deepcopy
import numpy as np
import os
from shutil import rmtree
import unittest
# Local imports
from . import nn_domains
from . import nn_modifiers
from .neural_network import NeuralNetwork
from .nn_visualise import visualise_nn
from .unittest_neural_network import generate_cnn_architectures, \
                                     generate_mlp_architectures
from ..utils.base_test_class import BaseTestClass, execute_tests


def networks_are_equal(net1, net2, false_if_net1_is_net2=True):
  """ Returns true if both net1 and net2 are equal.
      If any part of net1 is copied onto net2, then the output will be false
      if false_if_net1_is_net2 is True (default).
  """
  is_true = True
  for key in net1.__dict__.keys():
    val1 = net1.__dict__[key]
    val2 = net2.__dict__[key]
    is_true = True
    if isinstance(val1, dict):
      if false_if_net1_is_net2:
        is_true = is_true and (val1 is not val2)
      for val_key in val1.keys():
        is_true = is_true and np.all(val1[val_key] == val2[val_key])
    elif hasattr(val1, '__iter__'):
      if false_if_net1_is_net2:
        is_true = is_true and (val1 is not val2)
      is_true = is_true and np.all(val1 == val2)
    else:
      is_true = is_true and val1 == val2
    if not is_true: # break here if necessary
      return is_true
  return is_true


def run_orig_vs_modifications(save_dir, save_prefix, old_nn,
                                   get_modifications, constraint_checker, write_result):
  """ Tests for the original network and the modifications. Also, visualises the networks.
  """
  visualise_nn(old_nn, os.path.join(save_dir, '%s_orig'%(save_prefix)))
  old_nn_copy = deepcopy(old_nn)
  # Get the modified networks.
  new_nns = get_modifications(old_nn)
  # Go through each new network.
  for new_idx, new_nn in enumerate(new_nns):
    assert isinstance(new_nn, NeuralNetwork)
    assert constraint_checker(new_nn)
    visualise_nn(new_nn, os.path.join(save_dir, '%s_%d'%(save_prefix, new_idx)))
  # Finally test if the networks have not changed.
  assert networks_are_equal_are_equal(old_nn, old_nn_copy)
  write_result('%s (%s):: #new-networks: %d.'%(
    save_prefix, old_nn.nn_class, len(new_nns)), 'test_result')


@unittest.skip
class NNModifierTestCase(BaseTestClass):
  """ Unit tests for the NNModifier class. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(NNModifierTestCase, self).__init__(*args, **kwargs)
    self.cnns = generate_cnn_architectures()
    self.mlps = generate_mlp_architectures()
    self.save_dir = '../scratch/unittest_modifier_class/'
    self.cnn_constraint_checker = nn_domains.CNNConstraintChecker(
      50, 4, np.inf, 4.0, 5, 5, 100, 8000, 8)
    self.mlp_constraint_checker = nn_domains.MLPConstraintChecker(
      50, 4, np.inf, 4.0, 5, 5, 100, 8000, 8)
    self.cnn_modifier = nn_modifiers.NNModifier(self.cnn_constraint_checker)
    self.mlp_modifier = nn_modifiers.NNModifier(self.mlp_constraint_checker)
    self.modifier_wo_cc = nn_modifiers.NNModifier(None)

  def _get_modifier_and_cc(self, nn):
    """ Returns modifier for the neural network nn."""
    if nn.nn_class == 'cnn':
      modifier = self.cnn_modifier
      constraint_checker = self.cnn_constraint_checker
    else:
      modifier = self.mlp_modifier
      constraint_checker = self.mlp_constraint_checker
    return modifier, constraint_checker

  def test_get_primitives(self):
    """ Test for the get_primitives_grouped_by_type method. """
    self.report('Testing get_primitives_grouped_by_type')
    test_nns = self.cnns + self.mlps
    primitives, _ = self.cnn_modifier.get_primitives_grouped_by_type(self.cnns[0])
    self.report('Types of primitives: %s'%(primitives.keys()), 'test_result')
    for idx, nn in enumerate(test_nns):
      nn_copy = deepcopy(nn)
      modifier, _ = self._get_modifier_and_cc(nn)
      primitives, _ = modifier.get_primitives_grouped_by_type(nn)
      report_str = '%d (%s n=%d,m=%d):: '%(idx, nn.nn_class, nn.num_layers,
                                           nn.get_total_num_edges())
      total_num_primitives = 0
      for _, list_or_prims in primitives.items():
        report_str += '%d, '%(len(list_or_prims))
        total_num_primitives += len(list_or_prims)
      report_str += 'tot=%d'%(total_num_primitives)
      self.report(report_str, 'test_result')
      assert networks_are_equal_are_equal(nn_copy, nn)

  def test_get_single_step_modifications(self):
    """ Tests single step modifications. """
    self.report('Testing single step modifications.')
    save_dir = os.path.join(self.save_dir, 'single_step')
    if os.path.exists(save_dir):
      rmtree(save_dir)
    # Now iterate through the test networks
    test_nns = self.cnns + self.mlps
    for idx, old_nn in enumerate(test_nns):
      save_prefix = str(idx)
      modifier, constraint_checker = self._get_modifier_and_cc(old_nn)
      if idx in [2, 12]:
        num_modifications = 'all'
      else:
        num_modifications = 'all'
      get_modifications = lambda arg_nn: modifier.get_single_step_modifications(
                                           arg_nn, num_modifications)
      run_orig_vs_modifications(save_dir, save_prefix, old_nn,
        get_modifications, constraint_checker, self.report)

  def test_multi_step_modifications(self):
    """ Tests multi step modifications. """
    num_steps = 4
    self.report('Testing %d-step modifications.'%(num_steps))
    num_modifications = 20
    save_dir = os.path.join(self.save_dir, 'multi_step_%d'%(num_steps))
    if os.path.exists(save_dir):
      rmtree(save_dir)
    # Now iterate through the test networks
    test_nns = self.cnns + self.mlps
    for idx, old_nn in enumerate(test_nns):
      save_prefix = str(idx)
      modifier, constraint_checker = self._get_modifier_and_cc(old_nn)
      get_modifications = lambda arg_nn: modifier.get_multi_step_modifications(
                                           arg_nn, num_steps, num_modifications)
      run_orig_vs_modifications(save_dir, save_prefix, old_nn,
        get_modifications, constraint_checker, self.report)

  def test_call(self):
    """ Tests the __call__ function with a single input of the modifier. """
    self.report('Testing the __call__ function with single input of the modifier.')
    num_modifications = 20
    num_steps_probs = [0.5, 0.25, 0.125, 0.075, 0.05]
    save_dir = os.path.join(self.save_dir, 'modifier_call_single')
    if os.path.exists(save_dir):
      rmtree(save_dir)
    test_nns = self.cnns + self.mlps
    for idx, old_nn in enumerate(test_nns):
      save_prefix = str(idx)
      modifier, constraint_checker = self._get_modifier_and_cc(old_nn)
      get_modifications = lambda arg_nn: modifier(arg_nn, num_modifications,
                                                  num_steps_probs)
      run_orig_vs_modifications(save_dir, save_prefix, old_nn,
        get_modifications, constraint_checker, self.report)

  def test_call_with_list(self):
    """ Tests the __call__ function with a single input of the modifier. """
    self.report('Testing the __call__ function with a list of inputs.')
    num_modifications = 40
    num_steps_probs = [0.5, 0.25, 0.125, 0.075, 0.05]
    save_dir = os.path.join(self.save_dir, 'modifier_call_list')
    if os.path.exists(save_dir):
      rmtree(save_dir)
    test_probs = [self.cnns, self.mlps, generate_mlp_architectures('class')]
    for idx, prob in enumerate(test_probs):
      save_prefix = str(idx)
      modifier = self.modifier_wo_cc
      modifications = modifier(prob, num_modifications, num_steps_probs)
      for new_idx, new_nn in enumerate(modifications):
        assert isinstance(new_nn, NeuralNetwork)
        visualise_nn(new_nn, os.path.join(save_dir, '%s_%d'%(save_prefix, new_idx)))
      self.report('With list of %d nns(%s):: #new-networks: %d.'%(
                   len(prob), prob[0].nn_class, len(modifications)), 'test_result')

if __name__ == '__main__':
  execute_tests()




"""
  Test cases for nn_gpy.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name

import numpy as np
import unittest
# Local imports
from .syn_nn_functions import syn_func1_common, cnn_syn_func1, mlp_syn_func1
from .nn_examples import generate_cnn_architectures, generate_mlp_architectures
from . import nn_gp
from ..utils.ancillary_utils import get_list_of_floats_as_str
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.option_handler import load_options

_TOL = 1e-5

# Create data ---------------------------------------------------------------------------
def gen_gp_test_data():
  """ Generates test data for the unit tests. """
  # CNNs
  cnns = generate_cnn_architectures()
  np.random.shuffle(cnns)
  X1_tr = cnns[:7]
  Y1_tr = np.array([cnn_syn_func1(x) for x in X1_tr])
  X1_te = cnns[7:]
  Y1_te = np.array([cnn_syn_func1(x) for x in X1_te])
  nn_type_1 = 'cnn'
  # MLP regression
  regmlps = generate_mlp_architectures('reg')
  np.random.shuffle(regmlps)
  X2_tr = regmlps[:5]
  Y2_tr = np.array([mlp_syn_func1(x) for x in X2_tr])
  X2_te = regmlps[5:]
  Y2_te = np.array([mlp_syn_func1(x) for x in X2_te])
  nn_type_2 = 'mlp-reg'
  # MLP classification
  classmlps = generate_mlp_architectures('class')
  np.random.shuffle(classmlps)
  X3_tr = classmlps[:5]
  Y3_tr = np.array([syn_func1_common(x) for x in X3_tr])
  X3_te = classmlps[5:]
  Y3_te = np.array([syn_func1_common(x) for x in X3_te])
  nn_type_3 = 'mlp-class'
  return [(X1_tr, Y1_tr, X1_te, Y1_te, nn_type_1),
          (X2_tr, Y2_tr, X2_te, Y2_te, nn_type_2),
          (X3_tr, Y3_tr, X3_te, Y3_te, nn_type_3)]

# Some utilities we will need for testing ------------------------------------------------
def build_nngp_with_dataset(dataset, kernel_type, num_coeffs, dist_type):
  """ Builds a GP using the training set in dataset. """
  mean_func = lambda x: np.array([np.median(dataset[1])] * len(x))
  noise_var = (dataset[1].std() ** 2)/20
  kernel_hyperparams = get_kernel_hyperparams(num_coeffs, kernel_type, dist_type)
  return nn_gp.NNGP(dataset[0], dataset[1], kernel_type, mean_func, noise_var,
                    dataset[-1], kernel_hyperparams, dist_type)

def get_kernel_hyperparams(num_coeffs, kernel_type, dist_type):
  """ Returns the kernel hyperparams for the unit-tests below. """
  # data
  mislabel_coeffs = [2.0, 2.0, 1.0, 1.0, 1.0]
  struct_coeffs = [0.25, 0.5, 1.0, 2.0, 4.0]
  # kernel params
  mislabel_coeffs = mislabel_coeffs[:num_coeffs]
  struct_coeffs = struct_coeffs[:num_coeffs]
  lp_betas = [1e-6] * num_coeffs
  emd_betas = [1] * num_coeffs
  lp_powers = [1] * num_coeffs
  emd_powers = [2] * num_coeffs
  if dist_type == 'lp':
    betas = lp_betas
    powers = lp_powers
  elif dist_type == 'emd':
    betas = emd_betas
    powers = emd_powers
  elif dist_type == 'lp-emd':
    betas = [j for i in zip(lp_betas, emd_betas) for j in i]
    powers = [j for i in zip(lp_powers, emd_powers) for j in i]
  else:
    raise ValueError('Unknown dist_type: %s.'%(dist_type))
  # Now construct the dictionary
  kernel_hyperparams = {}
  kernel_hyperparams['mislabel_coeffs'] = mislabel_coeffs
  kernel_hyperparams['struct_coeffs'] = struct_coeffs
  kernel_hyperparams['betas'] = betas
  kernel_hyperparams['powers'] = powers
  kernel_hyperparams['non_assignment_penalty'] = 1.0
  if kernel_type in ['lpemd_prod', 'lp', 'emd']:
    kernel_hyperparams['scale'] = 1.0
  elif kernel_type in ['lpemd_sum']:
    kernel_hyperparams['alphas'] = [1.0, 1.0]
  return kernel_hyperparams

def fit_nngp_with_dataset(dataset, kernel_type, dist_type):
  """ Fits an NNGP to this dataset. """
  options = load_options(nn_gp.nn_gp_args, '')
  options.kernel_type = kernel_type
  options.dist_type = dist_type
  gp_fitter = nn_gp.NNGPFitter(dataset[0], dataset[1], dataset[-1],
                               options=options, reporter=None)
  fitted_result = gp_fitter.fit_gp()
  fitted_gp = fitted_result[1]
  return fitted_gp

def compute_average_prediction_error(dataset, preds, true_labels_idx=None):
  """ Computes the prediction error. """
  true_labels_idx = 3 if true_labels_idx is None else true_labels_idx
  return (np.linalg.norm(dataset[true_labels_idx] - preds)**2)/ \
          len(dataset[true_labels_idx])

# Test classes =========================================================================
@unittest.skip
class NNGPTestCase(BaseTestClass):
  """ Contains unit tests for the TransportNNDistanceComputer class. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(NNGPTestCase, self).__init__(*args, **kwargs)
    self.datasets = gen_gp_test_data()
    self.dist_types = ['lp', 'emd', 'lp-emd']
    self.kernel_types = {'lp': ['lp'], 'emd':['emd'],
                         'lp-emd':['lpemd_prod', 'lpemd_sum']}

  def test_basics(self):
    """ Tests for adding data, evaluation and marginal likelihood. """
    self.report('Testing adding data, evaluation and marginal likelihood.' +
                ' Probabilistic test, might fail.')
    num_coeffs_vals = [2, 1, 4, 5] * 5
    num_tests = 0
    num_successes = 0
    for dataset in self.datasets:
      for dist_type in self.dist_types:
        for kernel_type in self.kernel_types[dist_type]:
          curr_num_coeffs = num_coeffs_vals.pop(0)
          curr_gp = build_nngp_with_dataset(dataset, kernel_type, curr_num_coeffs,
                                            dist_type)
          # Predictions & Marginal likelihood
          curr_preds, _ = curr_gp.eval(dataset[2], 'std')
          curr_gp_err = compute_average_prediction_error(dataset, curr_preds)
          const_err = compute_average_prediction_error(dataset, dataset[1].mean())
          lml = curr_gp.compute_log_marginal_likelihood()
          is_success = curr_gp_err < const_err
          num_tests += 1
          num_successes += is_success
          self.report(('(%s, ntr=%d, nte=%d):: GP-lml=%0.4f, GP-err=%0.4f, ' +
                       'Const-err=%0.4f.  succ=%d')%(dataset[-1][:5], len(dataset[0]),
                       len(dataset[2]), lml, curr_gp_err, const_err, is_success),
                       'test_result')
    succ_frac = num_successes / float(num_tests)
    self.report('Summary: num_successes / num_floats = %d/%d = %0.4f'%(num_successes,
                num_tests, succ_frac), 'test_result')
    assert succ_frac > 0.5

  def test_hallucinated_predictions(self):
    """ Testing hallucinated predictions for NNGP. """
    self.report('Testing hallucinated predictions for NNGP.')
    num_coeffs_vals = [2, 1, 4, 5] * 5
    for dataset in self.datasets:
      for dist_type in self.dist_types:
        for kernel_type in self.kernel_types[dist_type]:
          curr_num_coeffs = num_coeffs_vals.pop(0)
          curr_gp = build_nngp_with_dataset(dataset, kernel_type, curr_num_coeffs,
                                            dist_type)
          curr_preds, curr_stds = curr_gp.eval(dataset[2][1:], 'std')
          ha_preds, ha_stds = curr_gp.eval_with_hallucinated_observations(
                                dataset[2][1:], dataset[0][4:] + dataset[2][:1], 'std')
          assert np.linalg.norm(curr_preds - ha_preds) < _TOL
          assert np.all(curr_stds >= ha_stds)


def return_kernel_hyperparams_as_str(kernel_hyperparams):
  """ Returns the kernel hyper-params as a string. """
  if 'scale' in kernel_hyperparams:
    scale_str = 'scale=%0.5f'%(kernel_hyperparams['scale'])
  else:
    scale_str = 'alphas=%s'%(get_list_of_floats_as_str(kernel_hyperparams['alphas'], 5))
  betas_str = 'betas=' + str(kernel_hyperparams['betas'])
  return scale_str + ', ' + betas_str


@unittest.skip
class NNGPFitterTestCase(BaseTestClass):
  """ Contains unit tests for the TransportNNDistanceComputer class. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(NNGPFitterTestCase, self).__init__(*args, **kwargs)
    self.datasets = gen_gp_test_data()
    self.dist_types = ['lp', 'emd', 'lp-emd']
    self.kernel_types = {'lp': ['lp'], 'emd':['emd'],
                         'lp-emd':['lpemd_prod', 'lpemd_sum']}

  @classmethod
  def _read_kernel_hyperparams(cls, kernel_hyperparams):
    """ Returns the kernel hyper-params as a string. """
    return return_kernel_hyperparams_as_str(kernel_hyperparams)

  def test_marg_likelihood_and_prediction(self):
    """ Tests marginal likelihood and prediction. """
    # pylint: disable=too-many-locals
    self.report('Testing evaluation and marginal likelihood on Fitted GP.' +
                ' Probabilistic test, might fail.')
    naive_num_coeffs_vals = [2, 1, 4, 5] * 5
    num_tests = 0
    num_lml_successes = 0
    num_const_err_successes = 0
    num_naive_err_successes = 0
    for dataset_idx, dataset in enumerate(self.datasets):
      for dist_type in self.dist_types:
        for kernel_type in self.kernel_types[dist_type]:
          # Build the naive GP
          curr_num_coeffs = naive_num_coeffs_vals.pop(0)
          naive_gp = build_nngp_with_dataset(dataset, kernel_type, curr_num_coeffs,
                                            dist_type)
          # Obtain a fitted GP
          fitted_gp = fit_nngp_with_dataset(dataset, kernel_type, dist_type)
          # Obtain log marginal likelihoods
          naive_lml = naive_gp.compute_log_marginal_likelihood()
          fitted_lml = fitted_gp.compute_log_marginal_likelihood()
          lml_succ = naive_lml < fitted_lml
          num_lml_successes += lml_succ
          self.report(('(%s, ntr=%d, nte=%d):: naive-gp-lml=%0.4f, ' +
                       'fitted-gp-lml=%0.4f, lml-succ=%d.')%(
                       dataset[-1][:5], len(dataset[0]), len(dataset[2]), naive_lml,
                       fitted_lml, lml_succ), 'test_result')
          # Predictions & Marginal likelihood
          naive_preds, _ = naive_gp.eval(dataset[2], 'std')
          naive_gp_err = compute_average_prediction_error(dataset, naive_preds)
          fitted_preds, _ = fitted_gp.eval(dataset[2], 'std')
          fitted_gp_err = compute_average_prediction_error(dataset, fitted_preds)
          const_err = compute_average_prediction_error(dataset, dataset[1].mean())
          const_err_succ = const_err > fitted_gp_err
          naive_err_succ = naive_gp_err > fitted_gp_err
          num_const_err_successes += const_err_succ
          num_naive_err_successes += naive_err_succ
          self.report(('  dataset #%d: const-err: %0.4f (%d), naive-gp-err=%0.4f (%d), ' +
                       'fitted-gp-err=%0.4f.')%(
                       dataset_idx, const_err, const_err_succ,
                       naive_gp_err, naive_err_succ, fitted_gp_err),
                      'test_result')
          # Print out betas
          self.report('  fitted kernel %s hyper-params: %s'%(kernel_type,
                      self._read_kernel_hyperparams(fitted_gp.kernel.hyperparams)),
                      'test_result')
          num_tests += 1
    # Print out some statistics
    lml_frac = float(num_lml_successes) / float(num_tests)
    const_err_frac = float(num_const_err_successes) / float(num_tests)
    naive_err_frac = float(num_naive_err_successes) / float(num_tests)
    self.report('LML tests success fraction = %d/%d = %0.4f'%(
                num_lml_successes, num_tests, lml_frac), 'test_result')
    self.report('Const-err success fraction = %d/%d = %0.4f'%(
                num_const_err_successes, num_tests, const_err_frac), 'test_result')
    self.report('Naive-GP-err success fraction = %d/%d = %0.4f'%(
                num_naive_err_successes, num_tests, naive_err_frac), 'test_result')
    assert num_lml_successes == num_tests
    assert const_err_frac >= 0.5
    assert naive_err_frac >= 0.3


if __name__ == '__main__':
  execute_tests(234234)




"""
  Unit tests for Random optimiser on Cartesian Product Domains.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=unused-argument

import os
import unittest
# Local imports
try:
  from ..test_data.park1_3.park1_3 import park1_3
  from ..test_data.park1_3.park1_3_mf import park1_3_mf
  from ..test_data.park1_3.park1_3_mf import cost as cost_park1_3_mf
  from ..test_data.park2_4.park2_4 import park2_4
  from ..test_data.park2_4.park2_4_mf import park2_4_mf
  from ..test_data.park2_4.park2_4_mf import cost as cost_park2_4_mf
  from ..exd.cp_domain_utils import get_processed_func_from_raw_func_for_cp_domain, \
                                  get_raw_point_from_processed_point, \
                                  load_cp_domain_from_config_file, \
                                  load_config_file
  from ..exd.experiment_caller import CPFunctionCaller, \
                                      get_multifunction_caller_from_config
  from ..exd.worker_manager import SyntheticWorkerManager
  from . import random_optimiser
  from ..utils.ancillary_utils import is_nondecreasing, get_list_of_floats_as_str
  from ..utils.base_test_class import BaseTestClass, execute_tests
  from ..utils.reporters import get_reporter
  RUN_TESTS = True
except ImportError:
  RUN_TESTS = False


@unittest.skipIf(not RUN_TESTS, "Unable to import submodules")
class CPOptimiserBaseTestCase(object):
  """ Base test class for Optimisers on Cartesian Product Spaces. """
  # pylint: disable=no-member

  def setUp(self):
    """ Set up. """
    self.max_capital = 30
    self._child_set_up()
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.opt_problems = [
      (test_data_pardir + '/test_data/park1_3/config.json', (park1_3,)),
      (test_data_pardir + '/test_data/park2_4/config.json', (park2_4,)),
      ]

  def _child_set_up(self):
    """ Child set up. """
    pass

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    raise NotImplementedError('Implement in a child class.')

  @classmethod
  def _run_optimiser(cls, raw_func, domain_config_file, worker_manager, max_capital, mode,
                     *args, **kwargs):
    """ Run the optimiser from given args. """
    raise NotImplementedError('Implement in a child class.')

  def test_instantiation(self):
    """ Tests instantiation of the optimiser. """
    self.report('Testing instantiation of optimiser.')
    for idx, (dcf, raw_prob_funcs) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing instantiation of optimiser for %s.'%(
                  idx + 1, len(self.opt_problems), dcf), 'test_result')
      config = load_config_file(dcf)
      if hasattr(config, 'fidel_space'):
        raw_func, raw_fidel_cost_func = raw_prob_funcs
      else:
        raw_func = raw_prob_funcs[0]
        raw_fidel_cost_func = None
      func_caller = get_multifunction_caller_from_config(raw_func, config,
                                  raw_fidel_cost_func=raw_fidel_cost_func)
      optimiser = self._child_instantiate_optimiser(func_caller, self.worker_manager_1,
                    options=None, reporter=get_reporter('silent'))
      self.report('Instantiated %s object.'%(type(optimiser)))
      for attr in dir(optimiser):
        if not attr.startswith('_'):
          self.report('optimiser.%s = %s'%(attr, str(getattr(optimiser, attr))),
                      'test_result')

  def _test_optimiser_results(self, opt_val, opt_point, history, dcf):
    """ Tests optimiser results. """
    cp_domain, orderings = load_cp_domain_from_config_file(dcf)
    raw_opt_point = get_raw_point_from_processed_point(opt_point,
                        cp_domain, orderings.index_ordering, orderings.dim_ordering)
    self.report('Opt point: proc=%s, raw=%s.'%(opt_point, raw_opt_point))
    self.report('True opt val sequence: %s.'%(
      get_list_of_floats_as_str(history.curr_true_opt_vals)))
    saved_in_history = [key for key, _ in list(history.__dict__.items()) if not
                        key.startswith('__')]
    self.report('Stored in history: %s.'%(saved_in_history), 'test_result')
    assert opt_val == history.curr_opt_vals[-1]
    assert len(opt_point) == cp_domain.num_domains
    assert is_nondecreasing(history.curr_opt_vals)
    assert is_nondecreasing(history.curr_true_opt_vals)

  def test_optimisation_single(self):
    """ Tests running of the optimiser. """
    self.report('')
    self.report('Testing optimisation on a single worker.')
    for idx, (dcf, raw_prob_funcs) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing optimisation  with 1 worker on %s.'%(
                  idx + 1, len(self.opt_problems), dcf), 'test_result')
      self.worker_manager_1.reset()
      opt_val, opt_point, history = self._run_optimiser(raw_prob_funcs, dcf,
                                      self.worker_manager_1, self.max_capital, 'asy')
      self._test_optimiser_results(opt_val, opt_point, history, dcf)
      self.report('')

  def test_optimisation_asynchronous(self):
    """ Tests running the optimiser in asynchronous mode. """
    self.report('')
    self.report('Testing optimisation on three asynchronous workers.')
    for idx, (dcf, raw_prob_funcs) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing optimisation with 3 asynchronous workers on %s.'%(
                   idx+1, len(self.opt_problems), dcf), 'test_result')
      self.worker_manager_3.reset()
      opt_val, opt_point, history = self._run_optimiser(raw_prob_funcs, dcf,
                              self.worker_manager_3, self.max_capital, 'asy')
      self._test_optimiser_results(opt_val, opt_point, history, dcf)
      self.report('')

  def test_optimisation_synchronous(self):
    """ Tests running the optimiser in asynchronous mode. """
    self.report('')
    self.report('Testing optimisation on three synchronous workers.')
    for idx, (dcf, raw_prob_funcs) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing optimisation with 3 synchronous workers on %s.'%(
                   idx+1, len(self.opt_problems), dcf), 'test_result')
      self.worker_manager_3.reset()
      opt_val, opt_point, history = self._run_optimiser(raw_prob_funcs, dcf,
                              self.worker_manager_3, self.max_capital, 'syn')
      self._test_optimiser_results(opt_val, opt_point, history, dcf)
      self.report('')


@unittest.skipIf(not RUN_TESTS, "Unable to import submodules")
class MFCPOptimiserBaseTestCase(CPOptimiserBaseTestCase):
  """ Base test class for Multi-fidelity Optimisers on Cartesian Product Spaces. """
  # pylint: disable=no-member

  def setUp(self):
    """ Set up. """
    self.max_capital = 20
    self._child_set_up()
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.opt_problems = [
      (test_data_pardir + '/test_data/park1_3/config_mf.json',
                                (park1_3_mf, cost_park1_3_mf)),
      (test_data_pardir + '/test_data/park2_4/config_mf.json',
                                (park2_4_mf, cost_park2_4_mf)),
      ]

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    raise NotImplementedError('Implement in a child class.')

  @classmethod
  def _run_optimiser(cls, raw_func, domain_config_file, worker_manager, max_capital, mode,
                     *args, **kwargs):
    """ Run the optimiser from given args. """
    raise NotImplementedError('Implement in a child class.')


@unittest.skipIf(not RUN_TESTS, "Unable to import submodules")
class CPRandomOptimiserTestCaseDefinitions(object):
  """ Define Unit tests for the Random optimiser on cartesian product spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    return random_optimiser.CPRandomOptimiser(func_caller, worker_manager,
                                              options=options, reporter=reporter)

  @classmethod
  def _run_optimiser(cls, raw_prob_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Run the optimiser from given args. """
    raw_func = raw_prob_funcs[0]
    cp_dom, orderings = load_cp_domain_from_config_file(domain_config_file)
    proc_func = get_processed_func_from_raw_func_for_cp_domain(
                  raw_func, cp_dom, orderings.index_ordering, orderings.dim_ordering)
    func_caller = CPFunctionCaller(proc_func, cp_dom, raw_func=raw_func,
                                   domain_orderings=orderings)
    return random_optimiser.random_optimiser_from_func_caller(func_caller, worker_manager,
                                                       max_capital, mode, *args, **kwargs)


@unittest.skipIf(not RUN_TESTS, "Unable to import submodules")
class CPRandomOptimiserTestCase(CPRandomOptimiserTestCaseDefinitions,
                                CPOptimiserBaseTestCase,
                                BaseTestClass):
  """ Unit tests for the Random optimiser on cartesian product spaces. """
  pass


@unittest.skipIf(not RUN_TESTS, "Unable to import submodules")
class MFCPRandomOptimiserTestCaseDefinitions(object):
  """ Unit tests for Multi-fidelity random optimiser on cartesian product spaces. """

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    return random_optimiser.MFCPRandomOptimiser(func_caller, worker_manager,
                                                options=options, reporter=reporter)

  @classmethod
  def _run_optimiser(cls, raw_prob_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Run the optimiser from given args. """
    raw_func, raw_fidel_cost_func = raw_prob_funcs
    config = load_config_file(domain_config_file)
    func_caller = get_multifunction_caller_from_config(raw_func, config,
                                raw_fidel_cost_func=raw_fidel_cost_func)
    return random_optimiser.mf_random_optimiser_from_func_caller(func_caller,
             worker_manager, max_capital, mode)


@unittest.skipIf(not RUN_TESTS, "Unable to import submodules")
class MFCPRandomOptimiserTestCase(MFCPRandomOptimiserTestCaseDefinitions,
                                  MFCPOptimiserBaseTestCase,
                                  BaseTestClass):
  """ Unit tests for Multi-fidelity random optimiser on cartesian product spaces. """
  pass


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for functions/classes in nn_constraint_checker.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import

# Local imports
from . import nn_domains
from .unittest_neural_network import generate_cnn_architectures, \
                                     generate_mlp_architectures
from ..utils.base_test_class import BaseTestClass, execute_tests


class NNConstraintCheckerTestCase(BaseTestClass):
  """ Contains unit tests for the TransportNNDistanceComputer class. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(NNConstraintCheckerTestCase, self).__init__(*args, **kwargs)
    self.nns = generate_cnn_architectures() + generate_mlp_architectures()
    self.cnn_constraint_checker = nn_domains.CNNConstraintChecker(
      25, 5, 500000, 0, 5, 2, 15, 512, 16, 4)
    self.mlp_constraint_checker = nn_domains.MLPConstraintChecker(
      25, 5, 500000, 900, 5, 2, 15, 30, 8)

  def test_constraint_checker(self):
    """ Tests if the constraints are satisfied for each network. """
    report_str = ('Testing constraint checker: max_layers=%d, max_mass=%d,' +
                  'max_out_deg=%d, max_in_deg=%d, max_edges=%d, max_2stride=%d.')%(
                  self.cnn_constraint_checker.max_num_layers,
                  self.cnn_constraint_checker.max_mass,
                  self.cnn_constraint_checker.max_in_degree,
                  self.cnn_constraint_checker.max_out_degree,
                  self.cnn_constraint_checker.max_num_edges,
                  self.cnn_constraint_checker.max_num_2strides,
                  )
    self.report(report_str)
    for nn in self.nns:
      if nn.nn_class == 'cnn':
        violation = self.cnn_constraint_checker(nn, True)
        constrain_satisfied = self.cnn_constraint_checker(nn)
        img_inv_sizes = [piis for piis in nn.post_img_inv_sizes if piis != 'x']
        nn_class_str = ', max_inv_size=%d '%(max(img_inv_sizes))
      else:
        violation = self.mlp_constraint_checker(nn, True)
        constrain_satisfied = self.mlp_constraint_checker(nn)
        nn_class_str = ' '
      self.report(('%s: #layers:%d, mass:%d, max_outdeg:%d, max_indeg:%d, ' +
                   '#edges:%d%s:: %s, %s')%(
                   nn.nn_class, len(nn.layer_labels), nn.get_total_mass(),
                   nn.get_out_degrees().max(), nn.get_in_degrees().max(),
                   nn.conn_mat.sum(), nn_class_str, str(constrain_satisfied), violation),
                   'test_result')
      assert (constrain_satisfied and violation == '') or \
             (not constrain_satisfied and violation != '')


if __name__ == '__main__':
  execute_tests()




"""
  Unittests for optimisers on CP domains with NN variables.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=abstract-method

import os
import unittest
# Local
from ..exd.worker_manager import SyntheticWorkerManager
from ..opt.unittest_cp_random_optimiser import CPOptimiserBaseTestCase, \
                                               CPRandomOptimiserTestCaseDefinitions, \
                                               MFCPRandomOptimiserTestCaseDefinitions
from ..opt.unittest_cp_ga_optimiser import CPGAOPtimiserTestCaseDefinitions
from ..opt.unittest_cp_gp_bandit import CPGPBanditTestCaseDefinitions
from ..opt.unittest_mf_cp_gp_bandit import MFCPGPBanditTestCaseDefinitions
from ..test_data.syn_cnn_2.syn_cnn_2 import syn_cnn_2
from ..test_data.syn_cnn_2.syn_cnn_2_mf import syn_cnn_2_mf
from ..test_data.syn_cnn_2.syn_cnn_2_mf import cost as cost_syn_cnn_2_mf
from ..utils.base_test_class import BaseTestClass, execute_tests


class NNCPOptimiserBaseTestCase(CPOptimiserBaseTestCase):
  """ Unit-tests Definitions for CP domains with NN variables. """
  # pylint: disable=no-init

  def setUp(self):
    """ Set up. """
    self.max_capital = 30
    self._child_set_up()
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.opt_problems = [
      (test_data_pardir + '/test_data/syn_cnn_2/config.json', (syn_cnn_2,)),
      ]


class MFNNCPOptimiserBaseTestCase(CPOptimiserBaseTestCase):
  """ Unit-tests Definitions for CP domains with NN variables. """
  # pylint: disable=no-init

  def setUp(self):
    """ Set up. """
    self.max_capital = 20
    self._child_set_up()
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.opt_problems = [
      (test_data_pardir + '/test_data/syn_cnn_2/config_mf.json',
                              (syn_cnn_2_mf, cost_syn_cnn_2_mf)),
      ]


# Random Optimiser Unit tests -----------------------------------------------------
@unittest.skip
class NNCPRandomOptimiserTestCase(CPRandomOptimiserTestCaseDefinitions,
                                  NNCPOptimiserBaseTestCase,
                                  BaseTestClass):
  """ Unit tests for the Random optimiser on cartesian product spaces. """
  pass


# @unittest.skip
class MFNNCPRandomOptimiserTestCase(MFCPRandomOptimiserTestCaseDefinitions,
                                    MFNNCPOptimiserBaseTestCase,
                                    BaseTestClass):
  """ Unit tests for Multi-fidelity random optimiser on cartesian product spaces. """
  pass


# GA Optimiser Unit tests -----------------------------------------------------------
@unittest.skip
class NNCPGAOPtimiserTestCase(CPGAOPtimiserTestCaseDefinitions,
                              NNCPOptimiserBaseTestCase,
                              BaseTestClass):
  """ Unit tests for GA optimiser on cartesian product spaces. """
  pass


# GPBandit Unit tests ---------------------------------------------------------------
@unittest.skip
class NNCPGPBanditTestCase(CPGPBanditTestCaseDefinitions,
                           NNCPOptimiserBaseTestCase,
                           BaseTestClass):
  """ Unit tests for GP Bandits on cartesian product spaces. """
  pass

@unittest.skip
class MFCPGPBanditTestCase(MFCPGPBanditTestCaseDefinitions,
                           MFNNCPOptimiserBaseTestCase,
                           BaseTestClass):
  """ Unit tests for GP Bandits on cartesian product spaces. """
  pass


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for metropolis sampling of continuous distributions.
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
import unittest

# pylint: disable=no-self-use

import numpy as np

# Local imports
from .continuous import Normal, Exponential, ContinuousUniform
from ..utils.base_test_class import BaseTestClass, execute_tests

class MetropolisTestCases(BaseTestClass):
  """ Unit tests for metropolis sampling """

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 100000
    self.threshold = 0.1

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples):
    """ Computes Mean """
    return np.mean(samples)

  def _compute_variance(self, samples):
    """ Computes Variance """
    return np.var(samples)

  def test_metropolis_normal(self):
    """ Tests metropolis sampling from Normal distribution """
    self.report('Test metropolis sampling from Normal Distribution.')
    mean = 11
    var = 3
    dist = Normal(mean, var)
    samples = dist.draw_samples('metropolis', self.size, np.array([11]))
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s :: test result: mean=%0.3f, variance=%0.3f'\
                 %(str(dist), mean_r, var_r), 'test_result')

  def test_metropolis_exponential(self):
    """ Tests metropolis sampling from Exponential distribution """
    self.report('Test metropolis sampling of Exponential Distribution.')
    lam = 2
    dist = Exponential(lam)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('metropolis', self.size, np.array([0.1]))
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert (samples >= 0).all()
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

  @unittest.skip('Stochastic')
  def test_metropolis_uniform(self):
    """ Tests metropolis sampling from Continuous Uniform """
    self.report('Test metropolis sampling of Continuous Uniform Distribution.')
    lower = -5
    upper = 5
    dist = ContinuousUniform(lower, upper)
    mean = dist.get_mean()
    var = dist.get_variance()
    samples = dist.draw_samples('metropolis', self.size, np.array([0.1]))
    mean_r = self._compute_mean(samples)
    var_r = self._compute_variance(samples)
    self._check_sample_sizes(samples)
    assert (samples >= lower).all() and (samples <= upper).all()
    assert abs(mean - mean_r) <= self.threshold
    assert abs(var - var_r) <= self.threshold
    self.report('%s, mean=%0.3f, variance=%0.3f :: test result: mean=%0.3f, '
                'variance=%0.3f'%(str(dist), mean, var, mean_r, var_r), 'test_result')

if __name__ == '__main__':
  execute_tests()
